import pymimir as mm
import random

from abc import ABC, abstractmethod

from .reward_functions import RewardFunction
from .trajectories import Trajectory


class SubtrajectorySampler(ABC):
    """
    Abstract base class for sampling subtrajectories.
    """
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def sample(self, state: mm.State, goal_condition: mm.GroundConjunctiveCondition) -> Trajectory | None:
        """Generate a subtrajectory from the state."""
        pass


class IWSubtrajectorySampler(SubtrajectorySampler):
    def __init__(self, reward_function: RewardFunction, width: int) -> None:
        super().__init__()
        self.reward_function = reward_function
        self.width = width

    def _get_goal_improvement(self,
                              state: mm.State,
                              achieved_goal_condition: list[mm.GroundLiteral],
                              unachieved_goal_condition: list[mm.GroundLiteral]) -> int:
        for literal in achieved_goal_condition:
            if not state.literal_holds(literal):
                return 0
        improvement = 0
        for literal in unachieved_goal_condition:
            if state.literal_holds(literal):
                improvement += 1
        return improvement

    def sample(self, state: mm.State, goal_condition: mm.GroundConjunctiveCondition) -> Trajectory | None:
        achieved_goal_condition: list[mm.GroundLiteral] = []
        unachieved_goal_condition: list[mm.GroundLiteral] = []
        for literal in goal_condition:
            if (state.literal_holds(literal)): achieved_goal_condition.append(literal)
            else: unachieved_goal_condition.append(literal)
        # Construct the search space graph.
        predecessors: dict[mm.State, tuple[mm.GroundAction, mm.State]] = {}
        distances: dict[mm.State, int] = {state: 0}
        def add_transition(current_state: mm.State, action: mm.GroundAction, _: float, successor_state: mm.State) -> None:
            nonlocal predecessors, distances
            current_distance = distances[current_state]
            successor_distance = current_distance + 1
            if (successor_state not in distances) or (successor_distance < distances[successor_state]):
                distances[successor_state] = successor_distance
                predecessors[successor_state] = (action, current_state)
        mm.iw(state.get_problem(), state, self.width, on_generate_new_state=add_transition)
        # Identify destination candidates.
        # We prioritize large overlaps and then short distances.
        best_improvement: int = 0
        best_distance: int = 0
        best_candidates: list[mm.State] = []
        for state, distance in distances.items():
            improvement = self._get_goal_improvement(state, achieved_goal_condition, unachieved_goal_condition)
            if improvement > best_improvement:
                best_improvement = improvement
                best_distance = distance
                best_candidates.clear()
                best_candidates.append(state)
            elif (improvement == best_improvement) and (distance < best_distance):
                best_candidates.clear()
                best_candidates.append(state)
            elif (improvement == best_improvement) and (distance == best_distance):
                best_candidates.append(state)
        # Uniformly sample a single candidate as the destination.
        # Then backtrack to the initial state to produce a subtrajectory.
        if (len(best_candidates) > 0) and (best_distance > 0):
            destination = random.choice(best_candidates)
            state_sequence = [destination]
            action_sequence = []
            value_sequence = []
            q_value_sequence = []
            reward_sequence = []
            state = destination
            while state in predecessors:
                action, predecessor_state = predecessors[state]
                reward = self.reward_function(predecessor_state, action, state, goal_condition)
                state_sequence.append(predecessor_state)
                action_sequence.append(action)
                value_sequence.append(float('nan'))  # TODO: Get a valid value using a model?
                q_value_sequence.append(float('nan'))  # TODO: Get a valid value using a model?
                reward_sequence.append(reward)
                state = predecessor_state
            state_sequence.reverse()
            action_sequence.reverse()
            value_sequence.reverse()
            q_value_sequence.reverse()
            reward_sequence.reverse()
            return Trajectory(state_sequence, action_sequence, value_sequence, q_value_sequence, reward_sequence, self.reward_function, goal_condition)
        # We were unable to find a single state that overlapped with the goal.
        return None



import pymimir as mm
import torch

from typing import Callable

from .goal_condition_sampling import GoalConditionSampler, OriginalGoalConditionSampler
from .initial_state_sampling import InitialStateSampler, OriginalInitialStateSampler
from .loss_functions import OptimizationFunction
from .problem_sampling import ProblemSampler, UniformProblemSampler
from .replay_buffers import ReplayBuffer
from .reward_functions import RewardFunction
from .trajectories import Trajectory, Transition
from .trajectory_refinements import TrajectoryRefiner, IdentityTrajectoryRefiner
from .trajectory_sampling import TrajectorySampler


class OffPolicyAlgorithm:
    """
    Modular off-policy RL algorithm.
    """

    def __init__(self,
                 problems: list[mm.Problem],
                 loss_function: OptimizationFunction,
                 reward_function: RewardFunction,
                 original_replay_buffer: ReplayBuffer,
                 hindsight_replay_buffer: ReplayBuffer,
                 trajectory_sampler: TrajectorySampler,
                 horizon: int,
                 rollout_count: int,
                 batch_size: int,
                 train_steps: int,
                 problem_sampler: ProblemSampler | None = None,
                 initial_state_sampler: InitialStateSampler |  None = None,
                 goal_condition_sampler: GoalConditionSampler |  None = None,
                 trajectory_refiner: TrajectoryRefiner |  None = None) -> None:
        """
        Initialize the off-policy RL algorithm with the specified components.

        Args:
            problems (list[mm.Problem]): List of problem instances.
            loss_function (LossFunction): Function to compute losses.
            reward_function (RewardFunction): Function to compute rewards.
            original_replay_buffer (ReplayBuffer): Buffer to store original experience transitions.
            hindsight_replay_buffer (ReplayBuffer): Buffer to store hindsight experience transitions.
            trajectory_sampler (TrajectorySampler): Sampler for generating trajectories.
            horizon (int): Maximum length of the sampled trajectories.
            rollout_count (int): Number of sampled trajectories.
            batch_size (int): Mini-batch size used for each optimization step.
            train_steps (int): Number of optimization steps to perform.
            problem_sampler (ProblemSampler, optional): Sampler for selecting problems. Defaults to UniformProblemSampler.
            initial_state_sampler (InitialStateSampler, optional): Sampler for initial states. Defaults to OriginalInitialStateSampler.
            goal_condition_sampler (GoalConditionSampler, optional): Sampler for goal conditions. Defaults to OriginalGoalConditionSampler.
            trajectory_refiner (TrajectoryRefiner, optional): Refiner for trajectories. Defaults to IdentityTrajectoryRefiner.
        """
        assert isinstance(problems, list) and all(isinstance(problem, mm.Problem) for problem in problems), "Problems must be a list of mm.Problem instances."
        assert len(problems) > 0, "At least one problem must be provided."
        assert isinstance(loss_function, OptimizationFunction), "Loss function must be an instance of LossFunction."
        assert isinstance(reward_function, RewardFunction), "Reward function must be an instance of RewardFunction."
        assert isinstance(hindsight_replay_buffer, ReplayBuffer), "Hindsight replay buffer must be an instance of ReplayBuffer."
        assert isinstance(original_replay_buffer, ReplayBuffer), "Original replay buffer must be an instance of ReplayBuffer."
        assert isinstance(trajectory_sampler, TrajectorySampler), "Trajectory sampler must be an instance of TrajectorySampler."
        assert isinstance(horizon, int), "Horizon must be an integer."
        assert horizon > 0, "Horizon must be positive."
        assert isinstance(train_steps, int), "Train steps must be an integer."
        assert train_steps > 0, "Train steps must be positive."
        self.problems = problems
        self.loss_function = loss_function
        self.reward_function = reward_function
        self.original_replay_buffer = original_replay_buffer
        self.hindsight_replay_buffer = hindsight_replay_buffer
        self.trajectory_sampler = trajectory_sampler
        self.horizon = horizon
        self.rollout_count = rollout_count
        self.batch_size = batch_size
        self.train_steps = train_steps
        self.problem_sampler = problem_sampler or UniformProblemSampler()
        self.initial_state_sampler = initial_state_sampler or OriginalInitialStateSampler()
        self.goal_condition_sampler = goal_condition_sampler or OriginalGoalConditionSampler()
        self.trajectory_refiner = trajectory_refiner or IdentityTrajectoryRefiner()
        assert isinstance(self.problem_sampler, ProblemSampler), "Problem sampler must be an instance of ProblemSampler."
        assert isinstance(self.initial_state_sampler, InitialStateSampler), "Initial state sampler must be an instance of InitialStateSampler."
        assert isinstance(self.goal_condition_sampler, GoalConditionSampler), "Goal condition sampler must be an instance of GoalConditionSampler."
        assert isinstance(self.trajectory_refiner, TrajectoryRefiner), "Trajectory refiner must be an instance of TrajectoryRefiner."
        # Initialize listener lists.
        self._listeners_sample_problems: list[Callable[[list[mm.Problem]], None]] = []
        self._listeners_sample_initial_states: list[Callable[[list[mm.State]], None]] = []
        self._listeners_sample_goal_conditions: list[Callable[[list[mm.GroundConjunctiveCondition]], None]] = []
        self._listeners_sample_trajectories: list[Callable[[list[Trajectory]], None]] = []
        self._listeners_refine_trajectories: list[Callable[[list[Trajectory]], None]] = []
        self._listeners_pre_collect_experience: list[Callable[[], None]] = []
        self._listeners_post_collect_experience: list[Callable[[], None]] = []
        self._listeners_train_step: list[Callable[[list[Transition], torch.Tensor], None]] = []
        self._listeners_pre_optimize_model: list[Callable[[], None]] = []
        self._listeners_post_optimize_model: list[Callable[[], None]] = []

    def _notify_sample_problems(self, result: list[mm.Problem]) -> None:
        for listener in self._listeners_sample_problems:
            listener(result)

    def _notify_sample_initial_states(self, result: list[mm.State]) -> None:
        for listener in self._listeners_sample_initial_states:
            listener(result)

    def _notify_sample_goal_conditions(self, result: list[mm.GroundConjunctiveCondition]) -> None:
        for listener in self._listeners_sample_goal_conditions:
            listener(result)

    def _notify_sample_trajectories(self, result: list[Trajectory]) -> None:
        for listener in self._listeners_sample_trajectories:
            listener(result)

    def _notify_refine_trajectories(self, result: list[Trajectory]) -> None:
        for listener in self._listeners_refine_trajectories:
            listener(result)

    def _notify_pre_collect_experience(self) -> None:
        for listener in self._listeners_pre_collect_experience:
            listener()

    def _notify_post_collect_experience(self) -> None:
        for listener in self._listeners_post_collect_experience:
            listener()

    def _notify_train_step(self, transitions: list[Transition], losses: torch.Tensor) -> None:
        for listener in self._listeners_train_step:
            listener(transitions, losses)

    def _notify_pre_optimize_model(self) -> None:
        for listener in self._listeners_pre_optimize_model:
            listener()

    def _notify_post_optimize_model(self) -> None:
        for listener in self._listeners_post_optimize_model:
            listener()

    def register_on_sample_problems(self, callback: Callable[[list[mm.Problem]], None]) -> None:
        """
        Register a callback to be called when problems are sampled in an episode.

        Args:
            callback (Callable[[list[mm.Problem]], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_sample_problems.append(callback)

    def register_on_sample_initial_states(self, callback: Callable[[list[mm.State]], None]) -> None:
        """
        Register a callback to be called when initial states are sampled in an episode.

        Args:
            callback (Callable[[list[mm.State]], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_sample_initial_states.append(callback)

    def register_on_sample_goal_conditions(self, callback: Callable[[list[mm.GroundConjunctiveCondition]], None]) -> None:
        """
        Register a callback to be called when goal conditions are sampled in an episode.

        Args:
            callback (Callable[[list[mm.GroundConjunctiveCondition]], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_sample_goal_conditions.append(callback)

    def register_on_sample_trajectories(self, callback: Callable[[list[Trajectory]], None]) -> None:
        """
        Register a callback to be called when trajectories are sampled in an episode.

        Args:
            callback (Callable[[list[Trajectory]], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_sample_trajectories.append(callback)

    def register_on_refine_trajectories(self, callback: Callable[[list[Trajectory]], None]) -> None:
        """
        Register a callback to be called when trajectories are refined in an episode.

        Args:
            callback (Callable[[list[Trajectory]], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_refine_trajectories.append(callback)

    def register_on_pre_collect_experience(self, callback: Callable[[], None]) -> None:
        """
        Register a callback to be called before collecting experience in an episode.

        Args:
            callback (Callable[[], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_pre_collect_experience.append(callback)

    def register_on_post_collect_experience(self, callback: Callable[[], None]) -> None:
        """
        Register a callback to be called after collecting experience in an episode.

        Args:
            callback (Callable[[], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_post_collect_experience.append(callback)

    def register_on_train_step(self, callback: Callable[[list[Transition], torch.Tensor], None]) -> None:
        """
        Register a callback to be called before optimizing the model parameters in an episode.

        Args:
            callback (Callable[[list[Transition], torch.Tensor], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_train_step.append(callback)

    def register_on_pre_optimize_model(self, callback: Callable[[], None]) -> None:
        """
        Register a callback to be called before optimizing the model parameters in an episode.

        Args:
            callback (Callable[[], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_pre_optimize_model.append(callback)

    def register_on_post_optimize_model(self, callback: Callable[[], None]) -> None:
        """
        Register a callback to be called after optimizing the model parameters in an episode.

        Args:
            callback (Callable[[], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_post_optimize_model.append(callback)

    def sample_problems(self, n: int) -> list[mm.Problem]:
        """
        Sample problems.

        Args:
            n (int): The number of problems to sample.

        Returns:
            list[mm.Problem]: A list of sampled problems.
        """
        result = self.problem_sampler.sample(self.problems, n)
        self._notify_sample_problems(result)
        return result

    def sample_initial_states(self, problems: list[mm.Problem]) -> list[mm.State]:
        """
        Sample initial states for the problems.

        Args:
            problems (list[mm.Problem]): A list of problem instances.

        Returns:
            list[mm.State]: A list of sampled initial states.
        """
        result = self.initial_state_sampler.sample(problems)
        self._notify_sample_initial_states(result)
        return result

    def sample_goal_conditions(self, problems: list[mm.Problem]) -> list[mm.GroundConjunctiveCondition]:
        """
        Sample goal conditions.

        Args:
            problems (list[mm.Problem]): A list of problem instances to sample goal conditions from.

        Returns:
            list[mm.GroundConjunctiveCondition]: A list of sampled goal conditions.
        """
        result = self.goal_condition_sampler.sample(problems)
        self._notify_sample_goal_conditions(result)
        return result

    def sample_trajectories(self, state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]]) -> list[Trajectory]:
        """
        Sample trajectories from the given problems.

        Args:
            state_goals (list[tuple[mm.State, mm.GroundConjunctiveCondition]]): A list of tuples containing states and their corresponding goals.
            model (ModelWrapper): A model used to sample successor states in the rollout process.
            reward_function (RewardFunction): A reward function to compute the rewards along the trajectory.

        Returns:
            list[Trajectory]: A list of sampled trajectories.
        """
        result = self.trajectory_sampler.sample(state_goals, self.horizon)
        self._notify_sample_trajectories(result)
        return result

    def refine_trajectories(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        """
        Refine the sampled trajectories.

        Args:
            trajectories (list[Trajectory]): A list of sampled trajectories.

        Returns:
            list[Trajectory]: A refined list of trajectories.
        """
        result = self.trajectory_refiner.refine(trajectories)
        self._notify_refine_trajectories(result)
        return result

    def fit(self) -> None:
        """
        Run one step of experience collection with refinement and model optimization.
        """
        self.collect_experience(self.rollout_count)
        self.optimize_model(self.batch_size)

    def collect_experience(self, rollout_size: int) -> None:
        """
        Collect experience from the sampled trajectories.

        Args:
            rollout_size (int): The number of trajectories to sample.
        """
        self._notify_pre_collect_experience()
        with torch.no_grad():
            problems = self.sample_problems(rollout_size)
            initial_states = self.sample_initial_states(problems)
            goal_conditions = self.sample_goal_conditions(problems)
            state_goals = list(zip(initial_states, goal_conditions))
            trajectories = self.sample_trajectories(state_goals)
            # Store successful trajectories in the original replay buffers.
            if self.original_replay_buffer is not None:
                for trajectory in trajectories:
                    if trajectory.is_solution():
                        for transition in trajectory:
                            self.original_replay_buffer.push(transition)
            # Store all relabeled trajectories in the hindsight replay buffer.
            if self.hindsight_replay_buffer is not None:
                refined_trajectories = self.refine_trajectories(trajectories)
                for refined_trajectory in refined_trajectories:
                    for refined_transition in refined_trajectory:
                        self.hindsight_replay_buffer.push(refined_transition)
        self._notify_post_collect_experience()

    def optimize_model(self, batch_size: int) -> None:
        """
        Update the strategy using a batch of data.

        Args:
            batch_size (int): The batch size for model optimization.
        """
        self._notify_pre_optimize_model()
        len_original = len(self.original_replay_buffer) if self.original_replay_buffer is not None else 0
        len_hindsight = len(self.hindsight_replay_buffer) if self.hindsight_replay_buffer is not None else 0
        buffer_size = len_original + len_hindsight
        if buffer_size > 0:
            for _ in range(self.train_steps):
                # Sample half the batch from the original buffer, if available.
                if self.original_replay_buffer is not None:
                    org_samples = min(len(self.original_replay_buffer), batch_size // 2)
                    org_transitions, org_weights, org_indices = self.original_replay_buffer.sample(org_samples)
                else:
                    org_samples = 0
                    org_transitions, org_weights, org_indices = [], torch.tensor([]), torch.tensor([])
                # Sample from the hindsight buffer to fill the rest of the batch, if available.
                if self.hindsight_replay_buffer is not None:
                    her_samples = batch_size - org_samples
                    her_transitions, her_weights, her_indices = self.hindsight_replay_buffer.sample(her_samples)
                else:
                    her_samples = 0
                    her_transitions, her_weights, her_indices = [], torch.tensor([]), torch.tensor([])
                # Combine buffers and compute losses.
                transitions = org_transitions + her_transitions
                weights = torch.cat((org_weights, her_weights), dim=0)
                losses = self.loss_function(transitions, weights)
                cpu_losses = losses.detach().cpu()
                if self.original_replay_buffer is not None:
                    self.original_replay_buffer.update(org_indices, cpu_losses[:org_samples])
                if self.hindsight_replay_buffer is not None:
                    self.hindsight_replay_buffer.update(her_indices, cpu_losses[org_samples:])
                self._notify_train_step(transitions, losses.detach())
        self._notify_post_optimize_model()



import pymimir as mm
import random

from abc import ABC, abstractmethod


class ProblemSampler(ABC):
    """
    Abstract base class for sampling problems.
    """

    @abstractmethod
    def sample(self, problems: list[mm.Problem], n: int) -> list[mm.Problem]:
        """
        Sample problems for the strategy.

        Args:
            problems (list[mm.Problem]): A list of problem instances.
            n (int): The number of problems to sample.

        Returns:
            list[mm.Problem]: A list of sampled problems.
        """
        pass


class UniformProblemSampler(ProblemSampler):
    """
    Samples the problems that are specified in the PDDL file.
    """

    def sample(self, problems: list[mm.Problem], n: int) -> list[mm.Problem]:
        return random.choices(problems, k=n)


from .algorithms import (
    OffPolicyAlgorithm
)

from .loss_functions import (
    DiscreteSoftActorCriticOptimization,
    DQNOptimization,
    OptimizationFunction
)

from .evaluations import (
    CoverageCriteria,
    EvaluationCriteria,
    LengthCriteria,
    PolicyEvaluation,
    ProbabilityCriteria,
    SequentialPolicyEvaluation,
    TDErrorCriteria
)

from .models import (
    ActionScalarModel
)

from .replay_buffers import (
    PrioritizedReplayBuffer,
    ReplayBuffer
)

from .reward_functions import (
    ConstantRewardFunction,
    FFRewardFunction,
    GoalTransitionRewardFunction,
    SumRewardFunction,
    RewardFunction
)

from .problem_sampling import (
    ProblemSampler,
    UniformProblemSampler
)

from .initial_state_sampling import (
    InitialStateSampler,
    OriginalInitialStateSampler,
    TopValueInitialStateSampler
)

from .goal_condition_sampling import (
    GoalConditionSampler,
    OriginalGoalConditionSampler
)

from .trajectory_sampling import (
    BeamSearchTrajectorySampler,
    BoltzmannTrajectorySampler,
    EpsilonGreedyTrajectorySampler,
    GreedyPolicyTrajectorySampler,
    PolicyTrajectorySampler,
    StateBoltzmannTrajectorySampler,
    TrajectorySampler
)

from .trajectories import (
    Trajectory,
    Transition
)

from .trajectory_refinements import (
    IdentityTrajectoryRefiner,
    LiftedHindsightTrajectoryRefiner,
    PropositionalHindsightTrajectoryRefiner,
    StateHindsightTrajectoryRefiner,
    PartialStateHindsightTrajectoryRefiner,
    TrajectoryRefiner
)

from .subtrajectory_sampling import (
    IWSubtrajectorySampler
)

# from .heuristics import (
#     LiftedFFHeuristic
# )

__all__ = [
    "ActionScalarModel",
    "BeamSearchTrajectorySampler",
    "BoltzmannTrajectorySampler",
    "ConstantRewardFunction",
    "CoverageCriteria",
    "DiscreteSoftActorCriticOptimization",
    "DQNOptimization",
    "EpsilonGreedyTrajectorySampler",
    "EvaluationCriteria",
    "FFRewardFunction",
    "GoalConditionSampler",
    "GoalTransitionRewardFunction",
    "GreedyPolicyTrajectorySampler",
    "IdentityTrajectoryRefiner",
    "InitialStateSampler",
    "IWSubtrajectorySampler",
    "LengthCriteria",
    "LiftedHindsightTrajectoryRefiner",
    "OffPolicyAlgorithm",
    "OptimizationFunction",
    "OriginalGoalConditionSampler",
    "OriginalInitialStateSampler",
    "PartialStateHindsightTrajectoryRefiner",
    "PolicyEvaluation",
    "PolicyTrajectorySampler",
    "PrioritizedReplayBuffer",
    "ProbabilityCriteria",
    "ProblemSampler",
    "PropositionalHindsightTrajectoryRefiner",
    "ReplayBuffer",
    "RewardFunction",
    "SequentialPolicyEvaluation",
    "StateBoltzmannTrajectorySampler",
    "StateHindsightTrajectoryRefiner",
    "SumRewardFunction",
    "TDErrorCriteria",
    "TopValueInitialStateSampler",
    "Trajectory",
    "TrajectoryRefiner",
    "TrajectorySampler",
    "Transition",
    "UniformProblemSampler",
    # "LiftedFFHeuristic",
]



import pymimir as mm
import torch


class ActionScalarModel(torch.nn.Module):
    """
    Abstract base class for model wrappers.
    """

    def forward(self, state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]]) -> list[tuple[torch.Tensor, list[mm.GroundAction]]]:
        """
        Forward pass through the model.

        Args:
            state_goals (list[State, GroundConjunctiveCondition]): The state-goal pairs to process.

        Returns:
            list[tuple[torch.Tensor, list[GroundAction]]]: For each state-goal pair, Q-values with corresponding valid actions.
        """
        raise NotImplementedError("Forward must be implemented by subclass.")



import math
import pymimir as mm
import torch

from abc import ABC, abstractmethod
from torch.nn.functional import huber_loss
from typing import Callable

from .models import ActionScalarModel
from .trajectories import Transition


"""
Abstract base class for optimization functions used in reinforcement learning.
"""
class OptimizationFunction(ABC):
    def get_value_bounds(self, transitions: list[Transition], device: torch.device) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Compute lower and upper bounds for values in the given transitions.

        Args:
            transitions (list[Transition]): List of Transition objects containing reward information.
            device (torch.device): PyTorch device to place the returned tensors on.

        Returns:
            Tuple of (lower_bounds_tensor, upper_bounds_tensor) containing the computed bounds for each transition.
        """
        lower_bounds: list[float] = []
        upper_bounds: list[float] = []
        for transition in transitions:
            reward_function = transition.reward_function
            lower_bound, upper_bound = reward_function.get_value_bounds(transition.immediate_reward, transition.future_rewards, transition.part_of_solution)
            lower_bounds.append(lower_bound)
            upper_bounds.append(upper_bound)
        lower_bounds_tensor = torch.tensor(lower_bounds, requires_grad=False, dtype=torch.float, device=device)
        upper_bounds_tensor = torch.tensor(upper_bounds, requires_grad=False, dtype=torch.float, device=device)
        return lower_bounds_tensor, upper_bounds_tensor

    @abstractmethod
    def __call__(self, transitions: list[Transition], weights: torch.Tensor) -> torch.Tensor:
        """
        Optimize the model parameters based on the given transitions and weights.

        Args:
            transitions (list[Transition]): List of Transition objects.
            weights (torch.Tensor): Tensor of weights for the transitions.

        Returns:
            Tensor containing the computed optimization errors. Gradient is not attached.
        """
        pass


class DQNOptimization(OptimizationFunction):

    def __init__(self,
                 model: ActionScalarModel,
                 model_optimizer: torch.optim.Optimizer,
                 model_lr_scheduler: torch.optim.lr_scheduler.LRScheduler,
                 target_model: ActionScalarModel,
                 discount_factor: float,
                 mellowmax_factor: float,
                 use_bounds_loss: bool = False) -> None:
        """
        Initializes the DQNOptimization class.

        Args:
            model (ActionScalarModel): The main model to be optimized.
            model_optimizer (torch.optim.Optimizer): The optimizer used for updating the model parameters.
            model_lr_scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler for the model optimizer.
            target_model (ActionScalarModel): The target model used for computing target values.
            discount_factor (float): The discount factor for future rewards. Must be in the range (0, 1].
            mellowmax_factor (float): The temperature parameter for the mellowmax function. Must be positive.
            use_bounds_loss (bool): If True, includes bounds loss in the optimization process. Defaults to False.

        Raises:
            AssertionError: If any of the arguments do not meet the specified type or value requirements.
        """
        assert isinstance(model, ActionScalarModel), "Model must be an instance of ActionScalarModel."
        assert isinstance(model_optimizer, torch.optim.Optimizer), "Model optimizer must be an instance of Optimizer."
        assert isinstance(model_lr_scheduler, torch.optim.lr_scheduler.LRScheduler), "Model learning rate scheduler must be an instance of LRScheduler."
        assert isinstance(target_model, ActionScalarModel), "Target model must be an instance of ActionScalarModel."
        assert isinstance(discount_factor, float), "Discount factor must be a float."
        assert isinstance(mellowmax_factor, float), "Mellowmax temperature must be a float."
        assert isinstance(use_bounds_loss, bool), "Option to use bounds loss must be a Boolean."
        assert discount_factor > 0.0, "Discount factor must be positive."
        assert discount_factor <= 1.0, "Discount factor must not be greater than 1."
        assert mellowmax_factor > 0.0, "Mellowmax temperature must be positive."
        self.model = model
        self.model_optimizer = model_optimizer
        self.model_lr_scheduler = model_lr_scheduler
        self.target_model = target_model
        self.discount_factor = discount_factor
        self.mellowmax_temperature = mellowmax_factor
        self.use_bounds_loss = use_bounds_loss

    def get_mellowmax_factor(self) -> float:
        """
        Get the mellowmax temperature factor.

        Returns:
            float: The current mellowmax temperature factor.
        """
        return self.mellowmax_temperature

    def set_mellowmax_factor(self, factor: float) -> None:
        """
        Set the mellowmax temperature factor.

        Args:
            factor (float): The new mellowmax temperature factor. Must be positive.
        """
        assert factor > 0.0, "Mellowmax temperature must be positive."
        self.mellowmax_temperature = factor

    def __call__(self, transitions: list[Transition], weights: torch.Tensor) -> torch.Tensor:
        """
        Execute the DQN optimization function.

        Args:
           transitions (list[Transition]): A list of transitions.
           weights (torch.Tensor): A tensor of weights corresponding to each transition.

        Returns:
           torch.Tensor: The loss tensor resulting from the optimization process, used for monitoring as no gradients are attached to it.
        """
        dead_end_value = -10000
        device = next(self.model.parameters()).device
        target_q_values = self._compute_targets(transitions, dead_end_value, device)
        state_goals = [(transition.current_state, transition.goal_condition) for transition in transitions]
        # Compute bounds, if selected.
        if self.use_bounds_loss:
            lower_bounds, upper_bounds = self.get_value_bounds(transitions, device)
        # Run the forward pass for each model, they share the target Q-values and bounds are only computed once.
        self.model.train()
        batch_q_values = self.model.forward(state_goals)
        selected_q_values = torch.stack([q_values[actions.index(transition.selected_action)] for (q_values, actions), transition in zip(batch_q_values, transitions)])
        # Compute DQN loss.
        losses = huber_loss(selected_q_values, target_q_values, delta=1.0, reduction='none')
        # Compute bounds loss, if selected.
        if self.use_bounds_loss:
            bounds_errors = selected_q_values - selected_q_values.clamp(lower_bounds, upper_bounds).detach()  # type: ignore
            bounds_losses = huber_loss(bounds_errors, torch.zeros_like(bounds_errors), delta=1.0, reduction='none')
            losses += bounds_losses
        losses *= weights.to(losses.device)
        self.model_optimizer.zero_grad()
        losses.mean().backward()
        self.model_optimizer.step()
        self.model_lr_scheduler.step()
        return losses.detach()

    def _mellowmax(self, values: torch.Tensor, dim: int = -1) -> torch.Tensor:
        """
        MellowMax operator (smooth maximum).

        Args:
            values (torch.Tensor): Tensor of values.
            dim (int): Dimension to apply the operator over (default: last).

        Returns:
            torch.Tensor: MellowMax values.
        """
        assert isinstance(values, torch.Tensor), "Values must be an instance of Tensor."
        return (torch.logsumexp(self.mellowmax_temperature * values, dim=dim) - torch.log(torch.tensor(float(values.size(dim)), device=values.device))) / self.mellowmax_temperature

    def _compute_targets(self, transitions: list[Transition], dead_end_value: float, device: torch.device) -> torch.Tensor:
        with torch.no_grad():
            successor_state_goals = [(transition.successor_state, transition.goal_condition) for transition in transitions]
            batch_successor_q_values = self.target_model.forward(successor_state_goals)
            successor_max_values = torch.stack([self._mellowmax(q_values) if (q_values.numel() > 0) else torch.tensor(dead_end_value, dtype=torch.float, device=device) for q_values, _ in batch_successor_q_values])
            immediate_rewards = torch.tensor([transition.immediate_reward for transition in transitions], requires_grad=False, dtype=torch.float, device=device)
            achieves_goal = torch.tensor([transition.achieves_goal for transition in transitions], dtype=torch.float, requires_grad=False, device=device)
            return immediate_rewards + (1.0 - achieves_goal) * self.discount_factor * successor_max_values


class DiscreteSoftActorCriticOptimization(OptimizationFunction):
    def __init__(self,
                 policy_model: ActionScalarModel,
                 policy_optimizer: torch.optim.Optimizer,
                 policy_lr_scheduler: torch.optim.lr_scheduler.LRScheduler,
                 qvalue_target_1: ActionScalarModel,
                 qvalue_model_1: ActionScalarModel,
                 qvalue_optimizer_1: torch.optim.Optimizer,
                 qvalue_lr_scheduler_1: torch.optim.lr_scheduler.LRScheduler,
                 qvalue_target_2: ActionScalarModel,
                 qvalue_model_2: ActionScalarModel,
                 qvalue_optimizer_2: torch.optim.Optimizer,
                 qvalue_lr_scheduler_2: torch.optim.lr_scheduler.LRScheduler,
                 discount_factor: float,
                 polyak_factor: float = 0.005,
                 entropy_target_scale: float = 1.0,
                 entropy_lr: float = 0.0003) -> None:
        """
        Initializes the DiscreteSoftActorCriticOptimization class.

        Args:
            policy_model (ActionScalarModel): The policy model to be optimized.
            policy_optimizer (torch.optim.Optimizer): The optimizer for the policy model.
            policy_lr_scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler for the policy optimizer.
            qvalue_target_1 (ActionScalarModel): The first Q-value model to compute targets.
            qvalue_model_1 (ActionScalarModel): The first Q-value model to be optimized.
            qvalue_optimizer_1 (torch.optim.Optimizer): The optimizer for the first Q-value model.
            qvalue_lr_scheduler_1 (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler for the first Q-value optimizer.
            qvalue_target_2 (ActionScalarModel): The second Q-value model to compute targets.
            qvalue_model_2 (ActionScalarModel): The second Q-value model to be optimized.
            qvalue_optimizer_2 (torch.optim.Optimizer): The optimizer for the second Q-value model.
            qvalue_lr_scheduler_2 (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler for the second Q-value optimizer.
            discount_factor (float): The discount factor for future rewards. Must be in the range (0, 1].
            polyak_factor (float): The rate at which the target models are updated. Must be in the range (0, 1]. Defaults to 0.005.
            entropy_target_scale (float): The initial entropy value. Must be in the range [0, 1]. Defaults to 1.0.
            entropy_lr (float): The learning rate for the entropy temperature. Must be positive. Defaults to 0.0003.

        Raises:
            AssertionError: If any of the arguments do not meet the specified type or value requirements.
        """
        super().__init__()
        assert isinstance(policy_model, ActionScalarModel), "Policy model must be an instance of ActionScalarModel."
        assert isinstance(policy_optimizer, torch.optim.Optimizer), "Policy optimizer must be an instance of torch.optim.Optimizer."
        assert isinstance(policy_lr_scheduler, torch.optim.lr_scheduler.LRScheduler), "Policy LR scheduler must be an instance of torch.optim.lr_scheduler.LRScheduler."
        assert isinstance(qvalue_target_1, ActionScalarModel), "First Q-value model must be an instance of ActionScalarModel."
        assert isinstance(qvalue_model_1, ActionScalarModel), "First Q-value model must be an instance of ActionScalarModel."
        assert isinstance(qvalue_optimizer_1, torch.optim.Optimizer), "First Q-value optimizer must be an instance of torch.optim.Optimizer."
        assert isinstance(qvalue_lr_scheduler_1, torch.optim.lr_scheduler.LRScheduler), "First Q-value LR scheduler must be an instance of torch.optim.lr_scheduler.LRScheduler."
        assert isinstance(qvalue_target_2, ActionScalarModel), "Second Q-value model must be an instance of ActionScalarModel."
        assert isinstance(qvalue_model_2, ActionScalarModel), "Second Q-value model must be an instance of ActionScalarModel."
        assert isinstance(qvalue_optimizer_2, torch.optim.Optimizer), "Second Q-value optimizer must be an instance of torch.optim.Optimizer."
        assert isinstance(qvalue_lr_scheduler_2, torch.optim.lr_scheduler.LRScheduler), "Second Q-value LR scheduler must be an instance of torch.optim.lr_scheduler.LRScheduler."
        assert isinstance(discount_factor, float), "Discount factor must be a float."
        assert isinstance(polyak_factor, float), "Polyak factor must be a float."
        assert isinstance(entropy_target_scale, float), "The initial entropy must be a float."
        assert isinstance(entropy_lr, float), "Learning rate for the entropy must be a float."
        assert discount_factor > 0.0, "Discount factor must be greater than 0.0."
        assert discount_factor <= 1.0, "Discount factor must be less than or equal to 1.0."
        assert polyak_factor > 0.0, "Polyak factor must be greater than 0.0."
        assert polyak_factor <= 1.0, "Polyak factor must be less than or equal to 1.0."
        assert entropy_target_scale >= 0.0, "Entropy factor must be greater than or equal to 0.0."
        assert entropy_lr > 0.0, "Learning rate for policy must be greater than 0.0."
        self.policy_model = policy_model
        self.policy_optimizer = policy_optimizer
        self.policy_lr_scheduler = policy_lr_scheduler
        self.qvalue_target_1 = qvalue_target_1
        self.qvalue_model_1 = qvalue_model_1
        self.qvalue_optimizer_1 = qvalue_optimizer_1
        self.qvalue_lr_scheduler_1 = qvalue_lr_scheduler_1
        self.qvalue_target_2 = qvalue_target_2
        self.qvalue_model_2 = qvalue_model_2
        self.qvalue_optimizer_2 = qvalue_optimizer_2
        self.qvalue_lr_scheduler_2 = qvalue_lr_scheduler_2
        self.discount_factor = discount_factor
        self.polyak_factor = polyak_factor
        device = next(self.policy_model.parameters()).device
        self.entropy_target_scale = entropy_target_scale
        self.log_entropy_alpha = torch.nn.Parameter(torch.tensor(0.0, device=device), requires_grad=True)
        self.entropy_optimizer = torch.optim.Adam([self.log_entropy_alpha], lr=entropy_lr)
        self._update_target_critics(1.0)  # Ensure that target models are initialized correctly.
        # Initialize listener lists.
        self._listeners_losses: list[Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], None]] = []

    def get_entropy_target_scale(self) -> float:
        return self.entropy_target_scale

    def set_entropy_target_scale(self, scale: float):
        assert scale >= 0.0, "Entropy factor must be greater than or equal to 0.0."
        self.entropy_target_scale = scale

    def get_entropy_alpha(self) -> float:
        return self.log_entropy_alpha.exp().item()

    def _notify_listeners_losses(self, actor_loss: torch.Tensor, critic_loss_1: torch.Tensor, critic_loss_2: torch.Tensor, entropy_loss: torch.Tensor):
        for listener in self._listeners_losses:
            listener(actor_loss, critic_loss_1, critic_loss_2, entropy_loss)

    def register_on_losses(self, callback: Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], None]):
        """
        Register a callback function to be called when the losses are computed.
        The callback function will be passed the actor loss, critic loss 1, critic loss 2, and entropy loss as arguments.

        Args:
           callback (Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], None]): The callback function to register.
        """
        assert callable(callback), "Callback must be a callable function."
        self._listeners_losses.append(callback)

    def _compute_qvalue_targets(self, transitions: list[Transition]) -> torch.Tensor:
        with torch.no_grad():
            device = next(self.policy_model.parameters()).device
            dead_end_value = -10000
            successor_state_goals = [(transition.successor_state, transition.goal_condition) for transition in transitions]
            batch_policy_logits = self.policy_model.forward(successor_state_goals)
            batch_qvalues_1 = self.qvalue_target_1.forward(successor_state_goals)
            batch_qvalues_2 = self.qvalue_target_2.forward(successor_state_goals)
            target_qvalues: list[torch.Tensor] = []
            for (policy_logits, _), (qvalues_1, _), (qvalues_2, _) in zip(batch_policy_logits, batch_qvalues_1, batch_qvalues_2):
                if policy_logits.numel() > 0:
                    entropy_alpha = self.log_entropy_alpha.exp().detach()
                    entropy_term = entropy_alpha * policy_logits.log_softmax(dim=-1)
                    min_qvalues = torch.min(qvalues_1, qvalues_2)
                    target_qvalue = (policy_logits.softmax(dim=-1) * (min_qvalues - entropy_term)).sum()
                    target_qvalues.append(target_qvalue)
                else:  # If there are no successor states, it's a dead end.
                    target_qvalues.append(torch.tensor(dead_end_value, device=device))
            immediate_rewards = torch.tensor([transition.immediate_reward for transition in transitions], requires_grad=False, dtype=torch.float, device=device)
            achieves_goal = torch.tensor([transition.achieves_goal for transition in transitions], dtype=torch.float, requires_grad=False, device=device)
            discounted_targets = immediate_rewards + (1.0 - achieves_goal) * self.discount_factor * torch.stack(target_qvalues)
            return discounted_targets

    def _compute_critic_losses(self,
                               transitions: list[Transition],
                               batch_qvalues_1: list[tuple[torch.Tensor, list[mm.GroundAction]]],
                               batch_qvalues_2: list[tuple[torch.Tensor, list[mm.GroundAction]]]) -> tuple[torch.Tensor, torch.Tensor]:
        qvalue_targets = self._compute_qvalue_targets(transitions)
        selected_qvalues_1 = torch.stack([qvalues[actions.index(transition.selected_action)] for (qvalues, actions), transition in zip(batch_qvalues_1, transitions)])
        selected_qvalues_2 = torch.stack([qvalues[actions.index(transition.selected_action)] for (qvalues, actions), transition in zip(batch_qvalues_2, transitions)])
        qvalue_losses_1 = torch.nn.functional.huber_loss(selected_qvalues_1, qvalue_targets, reduction='none')
        qvalue_losses_2 = torch.nn.functional.huber_loss(selected_qvalues_2, qvalue_targets, reduction='none')
        return qvalue_losses_1, qvalue_losses_2

    def _compute_actor_loss(self,
                            batch_qvalues_1: list[tuple[torch.Tensor, list[mm.GroundAction]]],
                            batch_qvalues_2: list[tuple[torch.Tensor, list[mm.GroundAction]]],
                            batch_policy_logits: list[tuple[torch.Tensor, list[mm.GroundAction]]]) -> torch.Tensor:
        batch_min_qvalues = [torch.min(qvalues_1, qvalues_2) for (qvalues_1, _), (qvalues_2, _) in zip(batch_qvalues_1, batch_qvalues_2)]
        batch_loss: list[torch.Tensor] = []
        for (policy_logits, _), min_qvalues in zip(batch_policy_logits, batch_min_qvalues):
            entropy_alpha = self.log_entropy_alpha.exp().detach()
            entropy_term = entropy_alpha * policy_logits.log_softmax(dim=-1)
            policy_loss = (policy_logits.softmax(dim=-1) * (entropy_term - min_qvalues.detach())).sum()
            batch_loss.append(policy_loss)
        policy_losses = torch.stack(batch_loss)
        return policy_losses

    # def _compute_entropy_loss(self, batch_policy_logits: list[tuple[torch.Tensor, list[mm.GroundAction]]]) -> torch.Tensor:
    #     batch_losses: list[torch.Tensor] = []
    #     for logits, _ in batch_policy_logits:
    #         entropy_policy = -(logits.softmax(dim=-1) * logits.log_softmax(dim=-1)).sum(0)
    #         entropy_target = self.entropy_target_scale * math.log(logits.numel())
    #         entropy_alpha = self.log_entropy_alpha.exp()
    #         entropy_loss = -entropy_alpha * (entropy_policy.detach() - entropy_target)
    #         batch_losses.append(entropy_loss)
    #     entropy_losses = torch.stack(batch_losses)
    #     return entropy_losses

    def _compute_entropy_loss(self, batch_policy_logits: list[tuple[torch.Tensor, list]]) -> torch.Tensor:
        losses = []
        for logits, _ in batch_policy_logits:
            probs = logits.softmax(dim=-1)
            log_probs = logits.log_softmax(dim=-1)
            entropy = -(probs * log_probs).sum(dim=-1)
            num_actions = logits.shape[-1]
            entropy_target = self.entropy_target_scale * math.log(num_actions)
            alpha_loss = self.log_entropy_alpha * (entropy.detach() - entropy_target)
            losses.append(alpha_loss)
        return torch.stack(losses)


    def _update_target_critics(self, polyak_factor: float) -> None:
        # Update target networks using polyak averaging.
        with torch.no_grad():
            for target_param, param in zip(self.qvalue_target_1.parameters(), self.qvalue_model_1.parameters()):
                target_param.copy_(polyak_factor * param + (1.0 - polyak_factor) * target_param)
            for target_param, param in zip(self.qvalue_target_2.parameters(), self.qvalue_model_2.parameters()):
                target_param.copy_(polyak_factor * param + (1.0 - polyak_factor) * target_param)

    def __call__(self, transitions: list[Transition], weights: torch.Tensor) -> torch.Tensor:
        """
        Execute the SAC optimization step.

        Args:
            transitions (list[Transition]): List of transitions.
            weights (torch.Tensor): Weights for the transitions.

        Returns:
            torch.Tensor: The loss tensor resulting from the optimization process, used for monitoring as no gradients are attached to it.
        """
        state_goals = [(transition.current_state, transition.goal_condition) for transition in transitions]
        batch_policy_logits = self.policy_model.forward(state_goals)
        batch_qvalues_1 = self.qvalue_model_1.forward(state_goals)
        batch_qvalues_2 = self.qvalue_model_2.forward(state_goals)

        # Update critics.
        critic_losses_1, critic_losses_2 = self._compute_critic_losses(transitions, batch_qvalues_1, batch_qvalues_2)

        self.qvalue_optimizer_1.zero_grad()
        critic_losses_1 *= weights.to(critic_losses_1.device)
        critic_losses_1.mean().backward()
        self.qvalue_optimizer_1.step()
        self.qvalue_lr_scheduler_1.step()

        self.qvalue_optimizer_2.zero_grad()
        critic_losses_2 *= weights.to(critic_losses_2.device)
        critic_losses_2.mean().backward()
        self.qvalue_optimizer_2.step()
        self.qvalue_lr_scheduler_2.step()

        # Update actor.
        actor_losses = self._compute_actor_loss(batch_qvalues_1, batch_qvalues_2, batch_policy_logits)

        self.policy_optimizer.zero_grad()
        actor_losses *= weights.to(actor_losses.device)
        actor_losses.mean().backward()
        self.policy_optimizer.step()
        self.policy_lr_scheduler.step()

        # Update entropy.
        entropy_losses = self._compute_entropy_loss(batch_policy_logits)

        self.entropy_optimizer.zero_grad()
        entropy_losses *= weights.to(entropy_losses.device)
        entropy_losses.mean().backward()
        self.entropy_optimizer.step()

        # Update critic targets.
        self._update_target_critics(self.polyak_factor)

        # Notify listeners about the loss values.
        self._notify_listeners_losses(actor_losses.detach(), critic_losses_1.detach(), critic_losses_2.detach(), entropy_losses.detach())

        # Return aggregated losses for monitoring.
        return (critic_losses_1 + critic_losses_2 + actor_losses + entropy_losses).detach()



import pymimir as mm

from abc import ABC, abstractmethod


class GoalConditionSampler(ABC):
    """
    Abstract base class for sampling goal conditions.
    """

    @abstractmethod
    def sample(self, problems: list[mm.Problem]) -> list[mm.GroundConjunctiveCondition]:
        """
        Sample goal conditions from the given problems.

        Args:
            problems (list[mm.Problem]): A list of problem instances.

        Returns:
            list[mm.GroundConjunctiveCondition]: A list of sampled goal conditions.
        """
        pass


class OriginalGoalConditionSampler(GoalConditionSampler):
    """
    Samples the goal condition that is specified in the PDDL file.
    """

    def sample(self, problems: list[mm.Problem]) -> list[mm.GroundConjunctiveCondition]:
        return [problem.get_goal_condition() for problem in problems]


# import pymimir as mm

# from pyswip import Prolog


# class LiftedFFHeuristic(mm.Heuristic):
#     def __init__(self, problem: mm.Problem, debug_mode: bool = False) -> None:
#         super().__init__()
#         self.problem = problem
#         self.debug_mode = debug_mode
#         self.rules = self._compute_rules()
#         self.goal = self._compute_query(self.problem.get_goal_condition())

#     def _to_prolog_term_name(self, name: str) -> str:
#         return name.removeprefix('?').upper() if name.startswith('?') else name.lower()

#     def _to_prolog_string(self, atom: mm.Atom | mm.GroundAtom) -> str:
#         predicate_name = atom.get_predicate().get_name()
#         term_names = [self._to_prolog_term_name(term.get_name()) for term in atom.get_terms()]
#         return f'{predicate_name}({str.join(', ', term_names)})'

#     def _compute_rules(self) -> list[str]:
#         rules: list[str] = []
#         for action in self.problem.get_domain().get_actions():
#             # Compute the body.
#             body = [self._to_prolog_string(literal.get_atom()) for literal in action.get_precondition().get_literals() if literal.get_polarity()]
#             # Compute the heads.
#             heads: list[str] = []
#             for conditional_effect in action.get_conditional_effect():
#                 assert len(conditional_effect.get_condition().get_parameters()) == 0, "Lifted FF does not support conditional effects with parameters."
#                 heads.extend([self._to_prolog_string(literal.get_atom()) for literal in conditional_effect.get_effect().get_literals() if literal.get_polarity()])
#             # Combine all heads with the body.
#             for head in heads:
#                 rules.append(f'{head} :- {str.join(', ', body)}')
#         return rules

#     def _compute_query(self, goal_condition: mm.GroundConjunctiveCondition) -> str:
#         body = [self._to_prolog_string(literal.get_atom()) for literal in goal_condition if literal.get_polarity()]
#         return f'has_relaxed_solution() :- {str.join(', ', body)}'

#     def compute_value(self, state: mm.State, is_goal_state: bool) -> float:
#         facts: list[str] = []
#         for atom in state.get_atoms():
#             predicate_name = atom.get_predicate().get_name().lower()
#             term_names = [obj.get_name() for obj in atom.get_terms()]
#             facts.append(f'{predicate_name}({str.join(', ', term_names)})')
#         # Write program to Prolog.
#         pl = Prolog()  # TODO: retractall to clear?
#         for fact in facts:
#             pl.assertz(fact, catcherrors=self.debug_mode)
#             print(fact)
#         for rule in self.rules:
#             pl.assertz(rule, catcherrors=self.debug_mode)
#             print(rule)
#         pl.assertz(self.goal)
#         print(self.goal)
#         results = list(pl.query('has_relaxed_solution()', maxresult=1))
#         pass
#         raise NotImplementedError()



import pymimir as mm

from .reward_functions import RewardFunction


class Transition:
    def __init__(self,
                 current_state: mm.State,
                 successor_state: mm.State,
                 selected_action: mm.GroundAction,
                 value: float,
                 q_value: float,
                 reward: float,
                 future_rewards: float,
                 reward_function: RewardFunction,
                 goal_condition: mm.GroundConjunctiveCondition,
                 part_of_solution: bool) -> None:
        assert current_state.get_problem() == successor_state.get_problem(), "Origin and destination states must belong to the same problem."
        self.current_state = current_state
        self.successor_state = successor_state
        self.selected_action = selected_action
        self.predicted_value = value
        self.predicted_q_value = q_value
        self.immediate_reward = reward
        self.future_rewards = future_rewards
        self.reward_function = reward_function
        self.goal_condition = goal_condition
        self.part_of_solution = part_of_solution
        self.achieves_goal = goal_condition.holds(successor_state)

    def __str__(self) -> str:
        return f'Transition({str(self.current_state)}, {str(self.selected_action)}, {str(self.successor_state)}, {str(self.goal_condition)})'


class Trajectory:
    def __init__(self,
                 state_sequence: list[mm.State],
                 action_sequence: list[mm.GroundAction],
                 value_sequence: list[float],
                 q_value_sequence: list[float],
                 reward_sequence: list[float],
                 reward_function: RewardFunction,
                 goal_condition: mm.GroundConjunctiveCondition):
        assert len(state_sequence) >= 1, "State sequence must contain at least one element."
        assert len(state_sequence) == len(action_sequence) + 1, "State sequence must have one more element than action sequence."
        assert len(action_sequence) == len(reward_sequence), "State sequence and reward sequence must have the same length."
        assert isinstance(goal_condition, mm.GroundConjunctiveCondition), "Goal condition must be a GroundConjunctiveCondition."
        self.problem = state_sequence[0].get_problem()
        self.start_state = state_sequence[0]
        self.final_state = state_sequence[-1]
        self.goal_condition = goal_condition
        self.achieves_goal = goal_condition.holds(self.final_state)
        self.reward_function = reward_function
        self.transitions: list[Transition] = []
        for idx in range(len(action_sequence)):
            current_state = state_sequence[idx]
            successor_state = state_sequence[idx + 1]
            selected_action = action_sequence[idx]
            value = value_sequence[idx]
            q_value = q_value_sequence[idx]
            reward = reward_sequence[idx]
            future_rewards = sum(reward_sequence[idx + 1:])
            transition = Transition(current_state, successor_state, selected_action, value, q_value, reward, future_rewards, reward_function, goal_condition, self.achieves_goal)
            self.transitions.append(transition)

    def __iter__(self):
        return iter(self.transitions)

    def __len__(self):
        return len(self.transitions)

    def __getitem__(self, index: int) -> Transition:
        return self.transitions[index]

    def is_solution(self) -> bool:
        return self.achieves_goal

    def clone_with_goal(self, start_index_incl: int, end_index_incl: int, goal_condition: mm.GroundConjunctiveCondition) -> 'Trajectory':
        assert start_index_incl >= 0 and end_index_incl < len(self.transitions), "Indices must be within the bounds of the trajectory."
        assert start_index_incl <= end_index_incl, "Start index must be less than or equal to end index."
        assert goal_condition.get_problem() == self.problem, "Goal condition must belong to the same problem as the trajectory."
        assert isinstance(goal_condition, mm.GroundConjunctiveCondition), "Goal condition must be a GroundConjunctiveCondition."
        cloned_state_sequence: list[mm.State] = []
        cloned_action_sequence: list[mm.GroundAction] = []
        cloned_value_sequence: list[float] = []
        cloned_q_value_sequence: list[float] = []
        cloned_reward_sequence: list[float] = []
        for transition_index in range(start_index_incl, end_index_incl + 1):
            transition = self.transitions[transition_index]
            current_state = transition.current_state
            selected_action = transition.selected_action
            successor_state = transition.successor_state
            reward = self.reward_function(current_state, selected_action, successor_state, goal_condition)
            cloned_state_sequence.append(current_state)
            cloned_action_sequence.append(selected_action)
            cloned_value_sequence.append(float("nan"))
            cloned_q_value_sequence.append(float("nan"))
            cloned_reward_sequence.append(reward)
        cloned_state_sequence.append(self.transitions[end_index_incl].successor_state)
        return Trajectory(cloned_state_sequence,
                          cloned_action_sequence,
                          cloned_value_sequence,
                          cloned_q_value_sequence,
                          cloned_reward_sequence,
                          self.reward_function,
                          goal_condition)

    def validate(self, should_achieve_goal: bool = True) -> None:
        """
        Validates the trajectory. Raises an AssertionError if any validation fails.
        """
        last_transition = self.transitions[-1]
        if should_achieve_goal:
            goal_condition = last_transition.goal_condition
            if goal_condition.holds(last_transition.successor_state):
                assert last_transition.achieves_goal
            if last_transition.achieves_goal:
                assert goal_condition.holds(last_transition.successor_state)
        for i in range(len(self.transitions) - 1):
            assert self.transitions[i].successor_state == self.transitions[i + 1].current_state
        for transition in self.transitions:
            assert isinstance(transition, Transition)
            assert isinstance(transition.goal_condition, mm.GroundConjunctiveCondition)
            assert transition.current_state.get_problem() == self.problem
            assert transition.successor_state.get_problem() == self.problem
            # actions = transition.current_state.generate_applicable_actions()
            # assert transition.selected_action in actions
            assert transition.selected_action.get_precondition().holds(transition.current_state)
            assert transition.selected_action.apply(transition.current_state) == transition.successor_state
            assert not transition.goal_condition.holds(transition.current_state)
        if len(self.transitions) > 0:
            expected_rewards = sum(transition.immediate_reward for transition in self.transitions)
            actual_rewards = self.transitions[0].immediate_reward + self.transitions[0].future_rewards
            assert actual_rewards == expected_rewards, "Expected and actual rewards must not differ."

    def __str__(self) -> str:
        return '[' + str.join(', ', [str(transition) for transition in self.transitions]) + ']'



import pymimir as mm
import random
import torch

from abc import ABC, abstractmethod
from collections import defaultdict

from .models import ActionScalarModel
from .reward_functions import RewardFunction
from .subtrajectory_sampling import SubtrajectorySampler
from .trajectories import Trajectory


class TrajectorySampler(ABC):
    """
    Abstract base class for sampling trajectories.
    """

    @abstractmethod
    def sample(self, initial_state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]], horizon: int) -> list[Trajectory]:
        """Generate trajectories for the given instances using the model."""


class PolicyRolloutSampler(TrajectorySampler):
    """
    Abstract base class for sampling trajectories.
    """
    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 subtrajectory_sampler: SubtrajectorySampler | None,
                 subtrajectory_sampling_probability: float) -> None:
        assert isinstance(model, ActionScalarModel), "Model must be an instance of ActionScalarModel."
        assert isinstance(reward_function, RewardFunction), "Reward function must be an instance of RewardFunction."
        assert isinstance(subtrajectory_sampler, SubtrajectorySampler) or (subtrajectory_sampler is None), "Subtrajectory sampler must be an instance of SubtrajectorySampler or None."
        assert isinstance(subtrajectory_sampling_probability, float), "Subtrajectory sampling probability must be a float."
        assert subtrajectory_sampling_probability >= 0.0, "Subtrajectory sampling probability must be at least 0.0."
        assert subtrajectory_sampling_probability <= 1.0, "Subtrajectory sampling probability must be at most 1.0."
        super().__init__()
        self.model = model
        self.reward_function = reward_function
        self.subtrajectory_sampler = subtrajectory_sampler
        self.subtrajectory_sampling_probability = subtrajectory_sampling_probability

    class RolloutContext:
        def __init__(self, initial_state: mm.State, goal_condition: mm.GroundConjunctiveCondition) -> None:
            self.state_sequence: list[mm.State] = [initial_state]
            self.action_sequence: list[mm.GroundAction] = []
            self.value_sequence: list[float] = []
            self.q_value_sequence: list[float] = []
            self.reward_sequence: list[float] = []
            self.closed_set: set[mm.State] = set([initial_state])
            self.goal_condition: mm.GroundConjunctiveCondition = goal_condition
            self.solved: bool = goal_condition.holds(initial_state)
            self.done = self.solved or (len(initial_state.generate_applicable_actions()) == 0)

    @abstractmethod
    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        """
        Sample an action index.

        Args:
            values (Tensor): A tensor of values, e.g., Q-values or logits.

        Returns:
            int: An index of an action.
        """
        pass

    def sample(self, initial_state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]], horizon: int) -> list[Trajectory]:
        """Generate trajectories for the given instances using the model."""
        rollout_contexts = [self.RolloutContext(initial_state, goal_condition) for initial_state, goal_condition in initial_state_goals]
        with torch.no_grad():
            self.model.eval()
            for _ in range(horizon):
                state_goals = [(context.state_sequence[-1], context.goal_condition) for context in rollout_contexts if not context.done]
                context_indices = [idx for idx, context in enumerate(rollout_contexts) if not context.done]
                if len(state_goals) == 0:
                    break  # All trajectories are done.
                q_values_batch = self.model.forward(state_goals)
                for rollout_idx, (q_values, applicable_actions) in zip(context_indices, q_values_batch):
                    context = rollout_contexts[rollout_idx]
                    rewards: list[float] = []
                    q_values = q_values.cpu()  # Move the result to CPU, the remaining operations are very cheap.
                    q_values_copy = q_values.clone()
                    current_state = context.state_sequence[-1]
                    # Reduce the value actions leading to already visited states.
                    for action_idx, action in enumerate(applicable_actions):
                        successor_state = action.apply(current_state)
                        rewards.append(self.reward_function(current_state, action, successor_state, context.goal_condition))
                        if successor_state in context.closed_set:
                            q_values_copy[action_idx] = -1000000.0
                    # Sample an action to apply.
                    action_idx = self.sample_action_index(current_state, applicable_actions, q_values_copy)
                    action = applicable_actions[action_idx]
                    successor_state = action.apply(current_state)
                    q_value = q_values[action_idx].item()
                    reward = rewards[action_idx]
                    value = (q_values + torch.tensor(rewards, dtype=torch.float, device=q_values.device)).max().item()
                    is_solved = context.goal_condition.holds(successor_state)
                    is_dead_end = len(successor_state.generate_applicable_actions()) == 0
                    exceeds_horizon = len(context.state_sequence) >= horizon
                    is_done = is_solved or is_dead_end or exceeds_horizon
                    # Update context.
                    context.state_sequence.append(successor_state)
                    context.action_sequence.append(action)
                    context.value_sequence.append(value)
                    context.q_value_sequence.append(q_value)
                    context.reward_sequence.append(reward)
                    context.closed_set.add(successor_state)
                    context.solved = is_solved
                    context.done = is_done
                # If enabled, sample a subtrajectory
                if (self.subtrajectory_sampler is not None) and (random.uniform(0.0, 1.0 - 1E-16) < self.subtrajectory_sampling_probability):
                    for rollout_idx in context_indices:
                        context = rollout_contexts[rollout_idx]
                        if not context.done:
                            start_state = context.state_sequence[-1]
                            goal_condition = context.goal_condition
                            subtrajectory = self.subtrajectory_sampler.sample(start_state, goal_condition)
                            if subtrajectory is not None:
                                for transition in subtrajectory:
                                    context.state_sequence.append(transition.successor_state)
                                    context.action_sequence.append(transition.selected_action)
                                    context.value_sequence.append(transition.predicted_value)
                                    context.q_value_sequence.append(transition.predicted_q_value)
                                    context.reward_sequence.append(transition.immediate_reward)
                                    context.closed_set.add(transition.successor_state)
                                final_state = context.state_sequence[-1]
                                is_solved = goal_condition.holds(final_state)
                                is_dead_end = len(final_state.generate_applicable_actions()) == 0
                                exceeds_horizon = len(context.state_sequence) >= horizon
                                is_done = is_solved or is_dead_end or exceeds_horizon
                                context.solved = is_solved
                                context.done = is_done
        # Create trajectories from contexts.
        trajectories: list[Trajectory] = []
        for context in rollout_contexts:
            trajectories.append(
                Trajectory(
                    context.state_sequence,
                    context.action_sequence,
                    context.value_sequence,
                    context.q_value_sequence,
                    context.reward_sequence,
                    self.reward_function,
                    context.goal_condition,
                )
            )
        return trajectories


class PolicyTrajectorySampler(PolicyRolloutSampler):
    """
    Values are treated as logits, and an action is sampled according to the resulting probability distribution.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 subtrajectory_sampler: SubtrajectorySampler | None = None,
                 subtrajectory_sampling_probability: float = 0.1) -> None:
        super().__init__(model, reward_function, subtrajectory_sampler, subtrajectory_sampling_probability)

    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        probabilities = values.softmax(0)
        action_index = probabilities.multinomial(num_samples=1)
        return action_index.item()  # type: ignore


class EpsilonGreedyTrajectorySampler(PolicyRolloutSampler):
    """
    Samples trajectories according to epsilon-greedy.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 epsilon: float,
                 subtrajectory_sampler: SubtrajectorySampler | None = None,
                 subtrajectory_sampling_probability: float = 0.1) -> None:
        assert isinstance(epsilon, float), "Epsilon must be a float."
        assert epsilon >= 0.0, "Epsilon must be a probability."
        assert epsilon <= 1.0, "Epsilon must be a probability."
        super().__init__(model, reward_function, subtrajectory_sampler, subtrajectory_sampling_probability)
        self.epsilon = epsilon

    def set_epsilon(self, epsilon: float) -> None:
        self.epsilon = epsilon

    def get_epsilon(self) -> float:
        return self.epsilon

    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        if torch.rand(1).item() < self.epsilon:
            return torch.randint(0, values.numel(), (1,)).item()  # type: ignore
        else:
            return torch.argmax(values).item()  # type: ignore


class BoltzmannTrajectorySampler(PolicyRolloutSampler):
    """
    Samples trajectories based on a Boltzmann distribution.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 temperature: float,
                 subtrajectory_sampler: SubtrajectorySampler | None = None,
                 subtrajectory_sampling_probability: float = 0.1) -> None:
        assert isinstance(temperature, float), "Temperature must be a float."
        assert temperature > 0.0, "Temperature must be positive."
        super().__init__(model, reward_function, subtrajectory_sampler, subtrajectory_sampling_probability)
        self.temperature = temperature

    def set_temperature(self, temperature: float) -> None:
        self.temperature = temperature

    def get_temperature(self) -> float:
        return self.temperature

    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        probabilities = (values / self.temperature).softmax(dim=0)
        action_index = probabilities.multinomial(num_samples=1)
        return action_index.item()  # type: ignore


class StateBoltzmannTrajectorySampler(PolicyRolloutSampler):
    """
    Samples trajectories based on a Boltzmann distribution, that takes into account state counts.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 initial_temperature: float,
                 final_temperature: float,
                 temperature_steps: int,
                 subtrajectory_sampler: SubtrajectorySampler | None = None,
                 subtrajectory_sampling_probability: float = 0.1) -> None:
        super().__init__(model, reward_function, subtrajectory_sampler, subtrajectory_sampling_probability)
        assert isinstance(initial_temperature, float), "Initial temperature must be a float."
        assert isinstance(final_temperature, float), "Final temperature must be a float."
        assert isinstance(temperature_steps, int), "Temperature steps must be an integer."
        assert initial_temperature > 0.0, "Initial temperature must be positive."
        assert final_temperature > 0.0, "Final temperature must be positive."
        assert temperature_steps > 0, "Temperature steps must be positive."
        self.initial_temperature = initial_temperature
        self.final_temperature = final_temperature
        self.temperature_steps = temperature_steps
        self.counts: defaultdict[mm.State, int] = defaultdict(int)

    def update_counts(self, trajectories: list[Trajectory]) -> None:
        # We do not count the final state in the state sequence, as no decision was taken in it.
        for trajectory in trajectories:
            for transition in trajectory:
                self.counts[transition.current_state] += 1

    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        state_counts = self.counts[state]
        ratio = min(1.0, state_counts / self.temperature_steps)
        temperature = self.initial_temperature * (1.0 - ratio) + self.final_temperature * ratio
        probabilities = (values / temperature).softmax(dim=0)
        action_idx = probabilities.multinomial(num_samples=1)
        return action_idx.item()  # type: ignore


class GreedyPolicyTrajectorySampler(PolicyRolloutSampler):
    """
    Samples trajectories based on greedily and deterministically following the policy.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 subtrajectory_sampler: SubtrajectorySampler | None = None,
                 subtrajectory_sampling_probability: float = 0.1) -> None:
        super().__init__(model, reward_function, subtrajectory_sampler, subtrajectory_sampling_probability)

    def sample_action_index(self, state: mm.State, actions: list[mm.GroundAction], values: torch.Tensor) -> int:
        action_index = values.argmax().item()
        return action_index  # type: ignore


class PolicySearchSampler(TrajectorySampler):
    """
    Abstract base class for sampling trajectories using search strategies.
    """

    @abstractmethod
    def sample(self, initial_state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]], horizon: int) -> list[Trajectory]:
        """Generate trajectories for the given instances using the model."""


# class GBFSTrajectorySampler(PolicySearchSampler):
#     """
#     Samples trajectories based on the Greedy Best-First Search (GBFS) strategy.
#     """

#     def __init__(self,
#                  model: ActionScalarModel,
#                  reward_function: RewardFunction) -> None:
#         super().__init__()
#         self.model = model
#         self.reward_function = reward_function

#     def sample(self, initial_state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]], horizon: int) -> list[Trajectory]:
#         """Generate trajectories for the given instances using the model."""
#         raise NotImplementedError("GBFSTrajectorySampler is not implemented yet.")


class BeamSearchTrajectorySampler(PolicySearchSampler):
    """
    Samples trajectories based on the Beam Search strategy.
    """

    def __init__(self,
                 model: ActionScalarModel,
                 reward_function: RewardFunction,
                 max_beam_size: int) -> None:
        super().__init__()
        self.model = model
        self.reward_function = reward_function
        self.max_beam_size = max_beam_size

    class SearchContext:
        def __init__(self, initial_state: mm.State, goal_condition: mm.GroundConjunctiveCondition) -> None:
            self.transition_map: dict[mm.State, tuple[mm.State, mm.GroundAction, float, float]] = {}
            self.value_map: dict[mm.State, float] = {}
            self.open_list: list[mm.State] = [initial_state]
            self.closed_set: set[mm.State] = set([initial_state])
            self.goal_condition: mm.GroundConjunctiveCondition = goal_condition
            self.solved: bool = goal_condition.holds(initial_state)
            self.done: bool = self.solved or (len(initial_state.generate_applicable_actions()) == 0)
            self.depth: int = 0
            if self.solved:
                self.value_map[initial_state] = 0.0

    def _beam_step(self, context: SearchContext, horizon: int, beam_successor_values: list[tuple[torch.Tensor, list[mm.GroundAction]]]) -> None:
        # Add current open list states to closed set.
        for state in context.open_list:
            context.closed_set.add(state)
        # Expand each beam.
        unique_successors = set()
        all_candidates: list[tuple[float, float, mm.State, mm.GroundAction, float, mm.State]] = []
        for open_idx, (successor_values, applicable_actions) in enumerate(beam_successor_values):
            successor_values = successor_values.cpu()
            priorities = successor_values.clone()
            current_state = context.open_list[open_idx]
            successor_states: list[mm.State] = []
            rewards: list[float] = []
            # Avoid actions leading to already visited states.
            for action_idx, action in enumerate(applicable_actions):
                successor_state = action.apply(current_state)
                successor_states.append(successor_state)
                rewards.append(self.reward_function(current_state, action, successor_state, context.goal_condition))
                if successor_state in context.closed_set:
                    priorities[action_idx] = -1000000.0
            # Add the value of the current state if not already present.
            if current_state not in context.value_map:
                current_value = (successor_values + torch.tensor(rewards, dtype=torch.float, device=successor_values.device)).max().item()
                context.value_map[current_state] = current_value
            # Sample an action to apply.
            for priority, successor_value, action, reward, successor_state in zip(priorities, successor_values, applicable_actions, rewards, successor_states):
                if (successor_state not in unique_successors) and (len(successor_state.generate_applicable_actions()) > 0):
                    unique_successors.add(successor_state)
                    all_candidates.append((priority.item(), successor_value.item(), current_state, action, reward, successor_state))
        # Select the most promising successors.
        all_candidates.sort(key=lambda x: x[0], reverse=True)
        best_candidates = all_candidates[:self.max_beam_size]
        # Update the beam context.
        for _, successor_value, current_state, action, reward, successor_state in best_candidates:
            if successor_state not in context.closed_set:
                context.transition_map[successor_state] = (current_state, action, reward, successor_value)
        context.open_list = [candidate[5] for candidate in best_candidates]
        context.solved = any(context.goal_condition.holds(state) for state in context.open_list)
        context.done = context.solved or (context.depth >= horizon) or (len(context.open_list) == 0)
        context.depth += 1

    def sample(self, initial_state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]], horizon: int) -> list[Trajectory]:
        """Generate trajectories for the given instances using the model."""
        contexts = [self.SearchContext(initial_state, goal_condition) for initial_state, goal_condition in initial_state_goals]
        with torch.no_grad():
            self.model.eval()
            for _ in range(horizon):
                state_goals_input: list[tuple[mm.State, mm.GroundConjunctiveCondition]] = []
                for context in contexts:
                    if not context.done:
                        state_goals_input.extend([(state, context.goal_condition) for state in context.open_list])
                if len(state_goals_input) == 0:
                    break  # All trajectories are done.
                # Evaluate all states in the beams.
                batch_successor_values = self.model.forward(state_goals_input)
                # Avoid visiting already visited states.
                offset = 0
                for context in contexts:
                    if not context.done:
                        beam_size = len(context.open_list)
                        beam_successor_values = batch_successor_values[offset:offset + beam_size]
                        self._beam_step(context, horizon, beam_successor_values)
                        offset += beam_size
        # Create trajectories from contexts.
        trajectories: list[Trajectory] = []
        for context in contexts:
            # Convert data to the trajectory format.
            state_sequence: list[mm.State] = []
            action_sequence: list[mm.GroundAction] = []
            reward_sequence: list[float] = []
            q_value_sequence: list[float] = []
            value_sequence: list[float] = []
            # Approximate the missing values of the final states by using the q-values.
            for state in context.open_list:
                if state not in context.value_map:
                    assert state in context.transition_map
                    _, _, _, successor_value = context.transition_map[state]
                    context.value_map[state] = successor_value
            # For each beam, produce a trajectory from the best state in the final open list.
            final_candidates = [(int(context.goal_condition.holds(state)), context.value_map[state], state) for state in context.open_list]
            final_candidates.sort(key=lambda x: (x[0], x[1]), reverse=True)
            _, value, state = final_candidates[0]
            state_sequence.append(state)
            value_sequence.append(value)
            while state in context.transition_map:
                predecessor_state, action, reward, q_value = context.transition_map[state]
                state_sequence.append(predecessor_state)
                action_sequence.append(action)
                reward_sequence.append(reward)
                q_value_sequence.append(q_value)
                value_sequence.append(value)
                state = predecessor_state
            state_sequence.reverse()
            action_sequence.reverse()
            reward_sequence.reverse()
            q_value_sequence.reverse()
            value_sequence.reverse()
            # Create trajectory.
            trajectories.append(
                Trajectory(
                    state_sequence,
                    action_sequence,
                    value_sequence,
                    q_value_sequence,
                    reward_sequence,
                    self.reward_function,
                    context.goal_condition,
                )
            )
        return trajectories



import math
import pymimir as mm

from abc import ABC, abstractmethod

from .trajectories import Trajectory
from .trajectory_sampling import TrajectorySampler


class EvaluationCriteria(ABC):
    @abstractmethod
    def evaluate(self, trajectories: list[Trajectory]) -> int:
        pass

    @abstractmethod
    def compare(self, x: int, y: int) -> int:
        """
        Compares two evaluation values.

        Args:
            x (int): The first evaluation value.
            y (int): The second evaluation value.

        Returns:
            int: -1 if x is worse than y, 0 if x and y are equal, and 1 if x is better than y.
        """
        pass


class CoverageCriteria(EvaluationCriteria):
    def evaluate(self, trajectories: list[Trajectory]) -> int:
        return sum([(1 if trajectory.is_solution() else 0) for trajectory in trajectories])

    def compare(self, x: int, y: int) -> int:
        return (x > y) - (x < y)


class LengthCriteria(EvaluationCriteria):
    def __init__(self, only_solutions: bool = True) -> None:
        super().__init__()
        self.only_solutions = only_solutions

    def evaluate(self, trajectories: list[Trajectory]) -> int:
        return sum([(len(trajectory) if (not self.only_solutions or trajectory.is_solution()) else 0) for trajectory in trajectories])

    def compare(self, x: int, y: int) -> int:
        return (x < y) - (x > y)


class TDErrorCriteria(EvaluationCriteria):
    def __init__(self, only_solutions: bool = True) -> None:
        super().__init__()
        self.only_solutions = only_solutions

    def evaluate(self, trajectories: list[Trajectory]) -> int:
        total_error = 0.0
        for trajectory in trajectories:
            if not self.only_solutions or trajectory.is_solution():
                for idx in range(len(trajectory) - 1):
                    curr = trajectory[idx]
                    succ = trajectory[idx + 1]
                    q_curr = curr.predicted_q_value
                    q_succ = succ.predicted_q_value + curr.immediate_reward
                    total_error += abs(q_succ - q_curr)
                total_error += abs(trajectory[-1].predicted_q_value)  # Terminal state should have Q-value 0.
        return round(total_error)

    def compare(self, x: int, y: int) -> int:
        return (x < y) - (x > y)


class ProbabilityCriteria(EvaluationCriteria):
    def __init__(self, only_solutions: bool = True) -> None:
        super().__init__()
        self.only_solutions = only_solutions

    @staticmethod
    def _sigmoid(x: float) -> float:
        return 1.0 / (1.0 + math.exp(-x))

    def evaluate(self, trajectories: list[Trajectory]) -> int:
        total_error = 0.0
        for trajectory in trajectories:
            if not self.only_solutions or trajectory.is_solution():
                for transition in trajectory:
                    # If the logit is negative, then the transition is undesirable and results in some error.
                    total_error += 1.0 - self._sigmoid(transition.predicted_q_value)
        return round(total_error)

    def compare(self, x: int, y: int) -> int:
        return (x < y) - (x > y)


class PolicyEvaluation:
    """
    Tracks and evaluates the performance of a policy across a set of problems using specified evaluation criteria.

    Args:
        problems (list[mm.Problem]): A list of problem instances to evaluate the policy on.
        criterias (list[EvaluationCriteria]): A list of evaluation criteria to assess policy performance.
        trajectory_sampler (TrajectorySampler): A specific strategy for generating trajectories with the model.
        reward_function (RewardFunction): A function for computing rewards along the generated trajectories.
        horizon (int): Maximum length of the generated trajectories.
    """
    def __init__(self,
                 problems: list[mm.Problem],
                 criterias: list[EvaluationCriteria],
                 trajectory_sampler: TrajectorySampler,
                 horizon: int) -> None:
        self.problems = problems
        self.criterias = criterias
        self.trajectory_sampler = trajectory_sampler
        self.horizon = horizon
        self.best_evaluation = None

    def evaluate(self) -> tuple[bool, list[int]]:
        """
        Evaluates the given model on the set of problems using the specified criteria.

        Args:
            model (ModelWrapper): The model to be evaluated.

        Returns:
            tuple[bool, list[int]]:
                - A boolean indicating whether the evaluation improved upon the best seen so far.
                - A list of evaluation scores for each criterion.
        """
        state_goals = [(problem.get_initial_state(), problem.get_goal_condition()) for problem in self.problems]
        trajectories = self.trajectory_sampler.sample(state_goals, self.horizon)
        # Evaluate on all trajectories.
        evaluation = [criteria.evaluate(trajectories) for criteria in self.criterias]
        if self.best_evaluation is None:
            self.best_evaluation = evaluation
            return True, evaluation
        # Compare evaluations lexicographically.
        for idx, criteria in enumerate(self.criterias):
            comparison = criteria.compare(evaluation[idx], self.best_evaluation[idx])
            if comparison < 0:
                return False, evaluation
            if comparison > 0:
                self.best_evaluation = evaluation
                return True, evaluation
        return False, evaluation


class SequentialPolicyEvaluation:
    """
    Tracks and evaluates the performance of a policy across a set of problems using specified evaluation criteria.
    In contrast to PolicyEvaluation, problems are evaluated sequentially in increasing order of difficulty, and evaluation stops as soon as one problem fails.
    The difficulty is determined by the size of the goal condition, and in case of ties, by the number of objects.
    The given criteria are only applied to the successfully solved problems.

    Args:
        problems (list[mm.Problem]): A list of problem instances to evaluate the policy on.
        criterias (list[EvaluationCriteria]): A list of evaluation criteria to assess policy performance.
        trajectory_sampler (TrajectorySampler): A specific strategy for generating trajectories with the model.
        reward_function (RewardFunction): A function for computing rewards along the generated trajectories.
        horizon (int): Maximum length of the generated trajectories.
    """
    def __init__(self,
                 problems: list[mm.Problem],
                 criterias: list[EvaluationCriteria],
                 trajectory_sampler: TrajectorySampler,
                 horizon: int) -> None:
        # Sort problems by difficulty: first by goal size, then by number of objects.
        self.problems = sorted(problems, key=lambda p: (len(p.get_goal_condition()), len(p.get_objects())))
        self.criterias = criterias
        self.trajectory_sampler = trajectory_sampler
        self.horizon = horizon
        self.best_evaluation = None

    def evaluate(self) -> tuple[bool, list[int]]:
        """
        Evaluates the given model on the set of problems using the specified criteria.

        Args:
            model (ModelWrapper): The model to be evaluated.

        Returns:
            tuple[bool, list[int]]:
                - A boolean indicating whether the evaluation improved upon the best seen so far.
                - A list of evaluation scores for each criterion.
        """
        # Generate trajectories for each problem sequentially.
        successful_trajectories = []
        for problem in self.problems:
            state_goal = (problem.get_initial_state(), problem.get_goal_condition())
            trajectory = self.trajectory_sampler.sample([state_goal], self.horizon)[0]
            if trajectory.is_solution():
                successful_trajectories.append(trajectory)
            else:
                break
        # Evaluate only on the successfully solved problems.
        evaluation = [criteria.evaluate(successful_trajectories) for criteria in self.criterias]
        if self.best_evaluation is None:
            self.best_evaluation = evaluation
            return True, evaluation
        # Compare evaluations lexicographically.
        for idx, criteria in enumerate(self.criterias):
            comparison = criteria.compare(evaluation[idx], self.best_evaluation[idx])
            if comparison < 0:
                return False, evaluation
            if comparison > 0:
                self.best_evaluation = evaluation
                return True, evaluation
        return False, evaluation



import torch

from abc import ABC, abstractmethod

from .trajectories import Transition


class ReplayBuffer(ABC):
    """
    Abstract base class for replay buffers.
    """

    @abstractmethod
    def push(self, transition: Transition) -> None:
        """
        Adds a transition to the replay buffer.
        """
        pass

    @abstractmethod
    def sample(self, n: int) -> tuple[list[Transition], torch.Tensor, torch.Tensor]:
        """
        Samples a batch of transitions from the replay buffer.

        Args:
            n (int): Number of transitions to sample.

        Returns:
            list[Transition]: A list of sampled transitions.
            torch.Tensor: Importance sampling weights.
            torch.Tensor: Indices of sampled transitions.
        """
        pass

    def update(self, indices: torch.Tensor, errors: torch.Tensor) -> None:
        """
        Updates the priorities of sampled transitions based on errors.

        Args:
            indices (torch.Tensor): Indices of sampled transitions.
            errors (torch.Tensor): Computed errors for sampled transitions.
        """
        pass

    @abstractmethod
    def __len__(self) -> int:
        """
        Returns the current number of stored transitions.
        """
        pass


class PrioritizedReplayBuffer(ReplayBuffer):
    """
    A Prioritized Experience Replay (PER) buffer that prioritizes sampling transitions
    based on their temporal difference (TD) error.

    Attributes:
        capacity (int): Maximum number of transitions the buffer can store.
        alpha (float): Controls the level of prioritization (0 = uniform sampling, 1 = full prioritization).
        beta (float): Controls importance sampling correction for bias reduction.
        beta_start (float): Initial beta value.
        beta_frames (int): Number of frames over which beta is annealed.
        buffer (list): Stores the experience tuples (state, action, reward, next_state, done).
        priorities (np.array): Stores the priorities of the transitions.
        position (int): Tracks the current position for overwriting old transitions.
        frame (int): Tracks the number of updates to anneal beta over time.
    """

    def __init__(self, capacity: int, alpha: float = 0.6, beta_start: float = 0.4, beta_frames: int = 100000):
        """
        Initializes the Prioritized Experience Replay buffer.

        Args:
            capacity (int): Maximum buffer size.
            alpha (float, optional): Degree of prioritization (default=0.6).
            beta_start (float, optional): Initial beta value for importance sampling (default=0.4).
            beta_frames (int, optional): Number of frames to anneal beta to 1.0 (default=100000).
        """
        self.capacity: int = capacity
        self.alpha: float = alpha
        self.beta_start: float = beta_start
        self.beta_frames: int = beta_frames
        self.buffer: list[Transition] = []
        self.position: int = 0
        self.priorities = torch.zeros(capacity, dtype=torch.float)
        self.beta: float = beta_start
        self.frame: int = 1  # Tracks number of updates for beta annealing.

    def push(self, transition: Transition) -> None:
        """
        Adds a transition to the replay buffer with the highest priority.
        Replaces the transition with the lowest priority when full.

        Args:
            state (np.array): The current state. action (int): The action taken.
            reward (float): The reward received. next_state (np.array): The next
            state. done (bool): Whether the episode ended.
        """
        max_priority = self.priorities.max() if self.buffer else 1.0  # New transitions get max priority
        if len(self.buffer) < self.capacity:
            self.buffer.append(transition)
        else:
            self.buffer[self.position] = transition
        self.priorities[self.position] = max_priority
        self.position = (self.position + 1) % self.capacity  # Circular buffer

    def sample(self, n: int) -> tuple[list[Transition], torch.Tensor, torch.Tensor]:
        """
        Samples a batch of transitions based on their priorities.

        Args:
            n (int): Number of transitions to sample.

        Returns:
            tuple: (transitions, weights, indices)
                - transitions (list[Transition]): Sampled transitions.
                - weights (np.array): Importance sampling weights.
                - indices (np.array): Indices of sampled transitions.
        """
        if len(self.buffer) == 0:
            return ([], torch.tensor([]), torch.tensor([]))
        # Compute probability distribution.
        priorities = self.priorities[:len(self.buffer)] ** self.alpha
        probabilities = priorities / priorities.sum()
        # Sample indices based on priority probabilities.
        indices = torch.multinomial(probabilities, n, replacement=True)
        samples: list[Transition] = [self.buffer[idx] for idx in indices]
        # Compute importance sampling (IS) weights.
        total = len(self.buffer)
        weights = (total * probabilities[indices]) ** -self.beta  # Normalize.
        weights /= weights.max()
        # Anneal beta towards 1.0.
        self.beta = min(1.0, self.beta_start + (1.0 - self.beta_start) * (self.frame / self.beta_frames))
        self.frame += 1
        # Return the sampled batch.
        return samples, weights, indices

    def update(self, indices: torch.Tensor, errors: torch.Tensor) -> None:
        """
        Updates the priorities of sampled transitions based on TD errors.

        Args:
            indices (np.array): Indices of sampled transitions.
            errors (np.array): Computed TD errors for sampled transitions.
        """
        if indices.numel() > 0:
            self.priorities[indices] = indices.abs() + 1e-5

    def get_mean_priority(self) -> float:
        if len(self.buffer) == 0: return 0.0
        return self.priorities[:len(self.buffer)].mean().item()

    def get_std_priority(self) -> float:
        if len(self.buffer) == 0: return 0.0
        return self.priorities[:len(self.buffer)].std().item()

    def __len__(self) -> int:
        """
        Returns the current number of stored transitions.

        Returns:
            int: Number of transitions in the buffer.
        """
        return len(self.buffer)



import itertools
import math
import networkx as nx
import pymimir as mm
import random

from abc import ABC, abstractmethod
from collections import defaultdict

from .trajectories import Trajectory


class TrajectoryRefiner(ABC):
    """
    Abstract base class for refining trajectories.
    """

    @abstractmethod
    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        """
        Refine the sampled trajectories.

        Args:
            trajectories (list[Trajectory]): A list of sampled trajectories.
            reward_function (RewardFunction): A reward function to compute the rewards along the refined trajectories.

        Returns:
            list[Trajectory]: A refined list of trajectories.
        """
        pass


class IdentityTrajectoryRefiner(TrajectoryRefiner):
    """
    Does not refine the trajectories.
    """

    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        return trajectories


class PartialStateHindsightTrajectoryRefiner(TrajectoryRefiner):
    def __init__(self, max_generated: int, remove_non_goal_predicates: bool = True) -> None:
        super().__init__()
        self.max_generated = max_generated
        self.remove_non_goal_predicates = remove_non_goal_predicates

    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        refined_trajectories: list[Trajectory] = []
        for trajectory in trajectories:
            if len(trajectory) == 0:
                refined_trajectories.append(trajectory)
            else:
                problem = trajectory.problem
                if self.remove_non_goal_predicates:
                    whitelist_predicates = set(literal.get_atom().get_predicate() for literal in problem.get_goal_condition())
                else:
                    whitelist_predicates = set(problem.get_domain().get_predicates())
                end_index = len(trajectory) - 1
                while end_index >= 0:
                    goal_atoms = [atom for atom in trajectory[end_index].successor_state.get_atoms() if atom.get_predicate() in whitelist_predicates]
                    if len(goal_atoms) > 0:
                        goal_size = random.randint(1, len(goal_atoms))
                        goal_atoms = random.sample(goal_atoms, goal_size)
                        goal_literals = [mm.GroundLiteral.new(atom, True, problem) for atom in goal_atoms]
                        goal_condition = mm.GroundConjunctiveCondition.new(goal_literals, problem)
                        closed_set: set[mm.State] = set()
                        start_index = end_index
                        while start_index >= 0:
                            transition = trajectory[start_index]
                            closed_set.add(transition.successor_state)
                            if (transition.current_state in closed_set) or goal_condition.holds(transition.current_state):
                                break
                            start_index -= 1
                        if (start_index + 1) <= end_index:
                            refined_trajectories.append(trajectory.clone_with_goal(start_index + 1, end_index, goal_condition))
                        if len(refined_trajectories) >= self.max_generated:
                            return refined_trajectories
                        end_index = start_index - 1
                    else:
                        end_index -= 1
        return refined_trajectories


class StateHindsightTrajectoryRefiner(TrajectoryRefiner):
    def __init__(self, max_generated: int):
        self.max_generated = max_generated

    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        refined_trajectories: list[Trajectory] = []
        for trajectory in trajectories:
            if len(trajectory) == 0:
                refined_trajectories.append(trajectory)
            else:
                end_index = len(trajectory) - 1
                while end_index >= 0:
                    goal_atoms = trajectory[end_index].successor_state.get_atoms()
                    goal_literals = [mm.GroundLiteral.new(atom, True, trajectory.problem) for atom in goal_atoms]
                    goal_condition = mm.GroundConjunctiveCondition.new(goal_literals, trajectory.problem)
                    closed_set: set[mm.State] = set()
                    start_index = end_index
                    while start_index >= 0:
                        transition = trajectory[start_index]
                        closed_set.add(transition.successor_state)
                        if (transition.current_state in closed_set) or goal_condition.holds(transition.current_state):
                            break
                        start_index -= 1
                    if (start_index + 1) <= end_index:
                        refined_trajectories.append(trajectory.clone_with_goal(start_index + 1, end_index, goal_condition))
                    if len(refined_trajectories) >= self.max_generated:
                        return refined_trajectories
                    end_index = start_index - 1
        return refined_trajectories


class PropositionalHindsightTrajectoryRefiner(TrajectoryRefiner):
    def __init__(self, problems: list[mm.Problem], max_generated: int):
        self.problems = problems
        self.max_generated = max_generated
        self.problem_subgoals: dict[mm.Problem, list[mm.GroundConjunctiveCondition]] = {}
        for problem in problems:
            goal_condition = problem.get_goal_condition()
            goal_literals: list[mm.GroundLiteral] = list(goal_condition)  # type: ignore
            # Enumerate all possible combinations of goal_atoms
            subgoals: list[mm.GroundConjunctiveCondition] = []
            for size in range(1, len(goal_literals) + 1):
                list_of_subgoal_literals = itertools.islice(itertools.combinations(goal_literals, size), 100)
                for subgoal_literals in list_of_subgoal_literals:
                    subgoal = mm.GroundConjunctiveCondition.new(list(subgoal_literals), problem)
                    subgoals.append(subgoal)
            subgoals.sort(key=len, reverse=True)
            self.problem_subgoals[problem] = subgoals

    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        refined_trajectories: list[Trajectory] = []
        for trajectory in trajectories:
            if len(trajectory) == 0:
                refined_trajectories.append(trajectory)
            else:
                assert trajectory.problem in self.problems
                subgoals = self.problem_subgoals[trajectory.problem]
                end_index = len(trajectory) - 1
                while end_index >= 0:
                    for subgoal in subgoals:
                        if subgoal.holds(trajectory[end_index].successor_state):
                            closed_set: set[mm.State] = set()
                            start_index = end_index
                            while start_index >= 0:
                                transition = trajectory[start_index]
                                closed_set.add(transition.successor_state)
                                if (transition.current_state in closed_set) or subgoal.holds(transition.current_state):
                                    break  # Stop when we have found the first "invalid" transition
                                start_index -= 1
                            if (start_index + 1) <= end_index:
                                refined_trajectories.append(trajectory.clone_with_goal(start_index + 1, end_index, subgoal))
                            end_index = start_index
                            break  # Only extract the largest subgoal per index
                    if len(refined_trajectories) >= self.max_generated:
                        return refined_trajectories
                    end_index -= 1
        return refined_trajectories


class LiftedHindsightTrajectoryRefiner(TrajectoryRefiner):
    def __init__(self, problems: list[mm.Problem], max_generated: int, max_subgoals_per_size: int = 2):
        self.problems = problems
        self.max_generated = max_generated
        self.problem_subgoals: dict[mm.Problem, list[mm.ConjunctiveCondition]] = {}
        self.problem_blacklist: dict[mm.Problem, list[mm.Predicate]] = {}
        for problem in problems:
            domain = problem.get_domain()
            subgoals: list[mm.ConjunctiveCondition] = []
            blacklist = [domain.get_predicate('=')] if domain.has_predicate('=') else []
            grounded_subgoals = self._generate_grounded_subgoals(problem)
            for subgoal_size in sorted(grounded_subgoals.keys(), reverse=True):
                selected_subgoals = grounded_subgoals[subgoal_size][:max_subgoals_per_size]
                subgoals.extend([subgoal.lift(True) for subgoal in selected_subgoals])
            # Warmup, ground() initializes the generator.
            initial_state = problem.get_initial_state()
            for subgoal in subgoals:
                subgoal.ground(initial_state, 1)
            self.problem_subgoals[problem] = subgoals
            self.problem_blacklist[problem] = blacklist

    def _generate_grounded_subgoals(self, problem: mm.Problem) -> dict[int, list[mm.GroundConjunctiveCondition]]:
        goal_literals: list[mm.GroundLiteral] = list(problem.get_goal_condition())  # type: ignore
        goal_graph = nx.Graph()
        # Add nodes.
        for idx_i in list(range(len(goal_literals))):
            goal_graph.add_node(idx_i)  # type: ignore
        # Add edges.
        for idx_i in range(len(goal_literals)):
            for idx_j in range(idx_i + 1, len(goal_literals)):
                objs_i = goal_literals[idx_i].get_atom().get_terms()
                objs_j = goal_literals[idx_j].get_atom().get_terms()
                if any(o in objs_j for o in objs_i):
                    goal_graph.add_edge(idx_i, idx_j)  # type: ignore
        # Enumerate all connected subcomponents.
        # The variable `all_subcomponents` contains a list of lists of connected subgraphs, one for each original connected component.
        MAX_SUBCOMPONENTS_PER_SIZE = 100
        all_subcomponents = []
        for component in nx.connected_components(goal_graph):  # type: ignore
            subcomponents = []
            for size in range(1, len(component) + 1):  # type: ignore
                size_count = 0
                for subset in itertools.combinations(component, size):  # type: ignore
                    if nx.is_connected(goal_graph.subgraph(subset)):  # type: ignore
                        subcomponents.append(subset)  # type: ignore
                        size_count += 1
                        if size_count >= MAX_SUBCOMPONENTS_PER_SIZE:
                            break
            all_subcomponents.append(subcomponents)  # type: ignore
        # Sample a number of combinations of all subcomponents.
        MAX_SUBGOAL_SIZE = 10
        MAX_SUBCOMPONENT_SAMPLES = 5
        MAX_COMBINATION_SAMPLES = 1000
        sampled_subgraphs = []
        total_subcomponents = len(all_subcomponents)  # type: ignore
        for num_sampled_subcomponents in range(1, min(MAX_SUBGOAL_SIZE, total_subcomponents) + 1):
            num_subcomponent_samples = min(MAX_SUBCOMPONENT_SAMPLES, math.comb(total_subcomponents, num_sampled_subcomponents))
            for sampled_subcomponent_indices in {tuple(sorted(random.sample(range(total_subcomponents), k=num_sampled_subcomponents))) for _ in range(num_subcomponent_samples)}:
                sampled_subcomponents = [all_subcomponents[index] for index in sampled_subcomponent_indices]  # type: ignore
                sampled_subgraphs.extend([sum(combination, ()) for combination in itertools.islice(itertools.product(*sampled_subcomponents), MAX_COMBINATION_SAMPLES)])  # type: ignore
        # Convert the combinations to actual grounded subgoals.
        all_subgoals = [
            mm.GroundConjunctiveCondition.new([goal_literals[i] for i in subgraph], problem)  # type: ignore
            for subgraph in sampled_subgraphs if len(subgraph) <= MAX_SUBGOAL_SIZE  # type: ignore
        ]
        subgoals_by_size = defaultdict(list)  # type: ignore
        for subgoal in all_subgoals:
            subgoals_by_size[len(subgoal)].append(subgoal)  # type: ignore
        return subgoals_by_size  # type: ignore

    def refine(self, trajectories: list[Trajectory]) -> list[Trajectory]:
        refined_trajectories: list[Trajectory] = []
        for trajectory in trajectories:
            assert trajectory.problem in self.problems
            subgoals = self.problem_subgoals[trajectory.problem]
            blacklist = self.problem_blacklist[trajectory.problem]
            end_index = len(trajectory) - 1
            while end_index >= 0:
                for lifted_subgoal in subgoals:
                    grounded_subgoals = lifted_subgoal.ground(trajectory[end_index].successor_state, 1, blacklist)
                    if len(grounded_subgoals) > 0:
                        grounded_subgoal =  grounded_subgoals[0]  # We only ask for one binding.
                        closed_set: set[mm.State] = set()
                        start_index = end_index
                        while start_index >= 0:
                            transition = trajectory[start_index]
                            closed_set.add(transition.successor_state)
                            if (transition.current_state in closed_set) or grounded_subgoal.holds(transition.current_state):
                                break  # Stop when we have found the first "invalid" transition
                            start_index -= 1
                        if (start_index + 1) <= end_index:
                            refined_trajectories.append(trajectory.clone_with_goal(start_index + 1, end_index, grounded_subgoal))
                        end_index = start_index
                        break  # Only extract the largest subgoal per index
                if len(refined_trajectories) >= self.max_generated:
                    return refined_trajectories
                end_index -= 1
        return refined_trajectories



import math
import pymimir as mm

from abc import ABC, abstractmethod
from collections import OrderedDict



class RewardFunction(ABC):
    @abstractmethod
    def __call__(
        self,
        current_state: mm.State,
        action: mm.GroundAction,
        successor_state: mm.State,
        goal_condition: mm.GroundConjunctiveCondition,
    ) -> float:
        pass

    @staticmethod
    def get_dead_end_reward() -> float:
        return -10_000

    def get_value_bounds(self, immediate_reward: float, future_rewards: float, part_of_solution: bool) -> tuple[float, float]:
        return float('-inf'), float('inf')


class SumRewardFunction(RewardFunction):
    def __init__(self, reward_functions: list[RewardFunction]) -> None:
        self.reward_functions = reward_functions

    def __call__(
        self,
        current_state: mm.State,
        action: mm.GroundAction,
        successor_state: mm.State,
        goal_condition: mm.GroundConjunctiveCondition,
    ) -> float:
        total_reward = 0.0
        for reward_function in self.reward_functions:
            total_reward += reward_function(current_state, action, successor_state, goal_condition)
        return total_reward

    def get_value_bounds(self, immediate_reward: float, future_rewards: float, part_of_solution: bool) -> tuple[float, float]:
        min_bound = float('inf')
        max_bound = float('-inf')
        for reward_function in self.reward_functions:
            local_min_bound, local_max_bound = reward_function.get_value_bounds(immediate_reward, future_rewards, part_of_solution)
            min_bound = min(min_bound, local_min_bound)
            max_bound = max(max_bound, local_max_bound)
        return min_bound, max_bound


class GoalTransitionRewardFunction(RewardFunction):
    def __init__(self, constant: float = 1.0) -> None:
        assert constant > 0.0, "Constant must be positive."
        self.constant = constant

    def __call__(
        self,
        current_state: mm.State,
        action: mm.GroundAction,
        successor_state: mm.State,
        goal_condition: mm.GroundConjunctiveCondition,
    ) -> float:
        return self.constant if goal_condition.holds(successor_state) else 0

    def get_value_bounds(self, immediate_reward: float, future_rewards: float, part_of_solution: bool) -> tuple[float, float]:
        return (0.0, self.constant) if part_of_solution else (float('-inf'), float('inf'))


class ConstantRewardFunction(RewardFunction):
    def __init__(self, constant: float = -1.0) -> None:
        assert constant < 0.0, "Constant must be negative."
        self.constant = constant

    def __call__(
        self,
        current_state: mm.State,
        action: mm.GroundAction,
        successor_state: mm.State,
        goal_condition: mm.GroundConjunctiveCondition,
    ) -> float:
        return self.constant

    def get_value_bounds(self, immediate_reward: float, future_rewards: float, part_of_solution: bool) -> tuple[float, float]:
        return (immediate_reward + future_rewards, self.constant) if part_of_solution else (float('-inf'), float('inf'))


class FFRewardFunction(RewardFunction):
    """
    Reward function based on the Fast-Forward (FF) heuristic.
    The reward is computed as the difference in FF heuristic values
    between the current state and the successor state.
    """

    def __init__(self) -> None:
        self.heuristics: dict[mm.Problem, mm.FFHeuristic] = {}
        self.cache: OrderedDict[tuple[mm.State, mm.GroundConjunctiveCondition], float] = OrderedDict()

    def __call__(
        self,
        current_state: mm.State,
        action: mm.GroundAction,
        successor_state: mm.State,
        goal_condition: mm.GroundConjunctiveCondition,
    ) -> float:
        # Obtain or create the FF heuristic for the problem.
        problem = current_state.get_problem()
        if problem not in self.heuristics:
            self.heuristics[problem] = mm.FFHeuristic(problem)
        heuristic = self.heuristics[problem]
        # Obtain the FF value for the current state.
        current_state_goal = (current_state, goal_condition)
        if current_state_goal in self.cache:
            ff_current = self.cache[current_state_goal]
        else:
            ff_current = heuristic.compute_value(current_state, goal_condition)
            self.cache[current_state_goal] = ff_current
        # Obtain the FF value for the successor state.
        successor_state_goal = (successor_state, goal_condition)
        if successor_state_goal in self.cache:
            ff_successor = self.cache[successor_state_goal]
        else:
            ff_successor = heuristic.compute_value(successor_state, goal_condition)
            self.cache[successor_state_goal] = ff_successor
        # Trim the cache to avoid excessive memory usage.
        while len(self.cache) > 10000:
            self.cache.popitem(last=False)
        # Return the difference in FF values as the reward, unless one of them is a dead-end state.
        reward = ff_current - ff_successor
        return reward if math.isfinite(reward) else self.get_dead_end_reward()

    def get_value_bounds(self, immediate_reward: float, future_rewards: float, part_of_solution: bool) -> tuple[float, float]:
        return (immediate_reward + future_rewards, float('inf')) if part_of_solution else (float('-inf'), float('inf'))


from typing import Callable
import pymimir as mm
import pytest
import torch

from pathlib import Path
from pymimir_rgnn import (
    RelationalGraphNeuralNetwork,
    HyperparameterConfig,
    ModuleConfig,
    SumAggregation,
    PredicateMLPMessages,
    MLPUpdates,
    StateEncoder,
    GroundActionsEncoder,
    GoalEncoder,
    ActionScalarDecoder
)
from pymimir_rl import *


TEST_DIR = Path(__file__).parent
DATA_DIR = TEST_DIR / 'data'


class RGNNWrapper(ActionScalarModel):
    def __init__(self, domain: mm.Domain) -> None:
        super().__init__()  # type: ignore

        hparam_config = HyperparameterConfig(
            domain=domain,
            num_layers=4,
            embedding_size=8
        )

        input_spec = (StateEncoder(), GroundActionsEncoder(), GoalEncoder())
        output_spec = [('q', ActionScalarDecoder(hparam_config))]

        module_config = ModuleConfig(
            aggregation_function=SumAggregation(),
            message_function=PredicateMLPMessages(hparam_config, input_spec),
            update_function=MLPUpdates(hparam_config)
        )

        self.rgnn = RelationalGraphNeuralNetwork(hparam_config, module_config, input_spec, output_spec)  # type: ignore

    def forward(self, state_goals: list[tuple[mm.State, mm.GroundConjunctiveCondition]]) -> list[tuple[torch.Tensor, list[mm.GroundAction]]]:
        input_list: list[tuple[mm.State, list[mm.GroundAction], mm.GroundConjunctiveCondition]] = []
        actions_list: list[list[mm.GroundAction]] = []
        for state, goal in state_goals:
            actions = state.generate_applicable_actions()
            input_list.append((state, actions, goal))
            actions_list.append(actions)
        q_values_list: list[torch.Tensor] = self.rgnn.forward(input_list).readout('q')  # type: ignore
        output = list(zip(q_values_list, actions_list))
        for tensor, _ in output:
            assert not tensor.isnan().any()
            assert not tensor.isinf().any()
        return output


def test_model_wrapper():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    current_state = problem.get_initial_state()
    goal_condition = problem.get_goal_condition()
    output = model.forward([(current_state, goal_condition)])
    assert output is not None
    assert isinstance(output, list)
    assert len(output) > 0
    assert isinstance(output[0][0], torch.Tensor)
    assert isinstance(output[0][1], list)
    assert len(output[0][1]) > 0


def test_dqn_loss():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    optimizer = torch.optim.Adam(model.parameters())
    lr_scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer)
    loss = DQNOptimization(model, optimizer, lr_scheduler, model, 0.999, 10.0, True)
    transitions: list[Transition] = []
    current_state = problem.get_initial_state()
    reward_function = ConstantRewardFunction(-1.0)
    for selected_action in current_state.generate_applicable_actions():
        successor_state = selected_action.apply(current_state)
        goal_condition = problem.get_goal_condition()
        reward = reward_function(current_state, selected_action, successor_state, goal_condition)
        transitions.append(Transition(current_state, successor_state, selected_action, -1.0, -1.0, reward, 0.0, reward_function, goal_condition, False))
    losses = loss(transitions, torch.ones(len(transitions)))
    assert losses is not None
    assert len(losses) == len(transitions)


def test_sac_loss():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    policy_model = RGNNWrapper(domain)
    qvalue_model_1 = RGNNWrapper(domain)
    qvalue_model_2 = RGNNWrapper(domain)
    qvalue_target_1 = RGNNWrapper(domain)
    qvalue_target_2 = RGNNWrapper(domain)
    policy_optimizer = torch.optim.Adam(policy_model.parameters())
    qvalue_optimizer_1 = torch.optim.Adam(qvalue_model_1.parameters())
    qvalue_optimizer_2 = torch.optim.Adam(qvalue_model_2.parameters())
    policy_lr_scheduler = torch.optim.lr_scheduler.StepLR(policy_optimizer, step_size=10, gamma=0.9)
    qvalue_lr_scheduler_1 = torch.optim.lr_scheduler.StepLR(qvalue_optimizer_1, step_size=10, gamma=0.9)
    qvalue_lr_scheduler_2 = torch.optim.lr_scheduler.StepLR(qvalue_optimizer_2, step_size=10, gamma=0.9)
    discount_factor = 0.999
    polyak_factor = 0.005
    entropy_temperature = 1.0
    entropy_lr = 0.0003
    loss = DiscreteSoftActorCriticOptimization(policy_model,
                                               policy_optimizer,
                                               policy_lr_scheduler,
                                               qvalue_target_1,
                                               qvalue_model_1,
                                               qvalue_optimizer_1,
                                               qvalue_lr_scheduler_1,
                                               qvalue_target_2,
                                               qvalue_model_2,
                                               qvalue_optimizer_2,
                                               qvalue_lr_scheduler_2,
                                               discount_factor,
                                               polyak_factor,
                                               entropy_temperature,
                                               entropy_lr)
    transitions: list[Transition] = []
    current_state = problem.get_initial_state()
    reward_function = ConstantRewardFunction(-1.0)
    for selected_action in current_state.generate_applicable_actions():
        successor_state = selected_action.apply(current_state)
        goal_condition = problem.get_goal_condition()
        reward = reward_function(current_state, selected_action, successor_state, goal_condition)
        transitions.append(Transition(current_state, successor_state, selected_action, -1.0, -1.0, reward, 0.0, reward_function, goal_condition, False))
    losses = loss(transitions, torch.ones(len(transitions)))
    assert losses is not None
    assert len(losses) == len(transitions)


def test_ff_reward_function():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    reward_function = FFRewardFunction()
    current_state = problem.get_initial_state()
    goal_condition = problem.get_goal_condition()
    for action in current_state.generate_applicable_actions():
        successor_state = action.apply(current_state)
        reward = reward_function(current_state, action, successor_state, goal_condition)
        assert isinstance(reward, float)


def test_sum_reward_function():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    reward_function_1 = ConstantRewardFunction(-1.0)
    reward_function_2 = FFRewardFunction()
    sum_reward_function = SumRewardFunction([reward_function_1, reward_function_2])
    current_state = problem.get_initial_state()
    goal_condition = problem.get_goal_condition()
    for action in current_state.generate_applicable_actions():
        successor_state = action.apply(current_state)
        reward = sum_reward_function(current_state, action, successor_state, goal_condition)
        assert isinstance(reward, float)


@pytest.mark.parametrize("domain_name, trajectory_sampler_creator", [
    ('blocks', lambda model, reward_function: PolicyTrajectorySampler(model, reward_function)),
    ('blocks', lambda model, reward_function: BoltzmannTrajectorySampler(model, reward_function, 1.0)),
    ('blocks', lambda model, reward_function: StateBoltzmannTrajectorySampler(model, reward_function, 1.0, 0.1, 10)),
    ('blocks', lambda model, reward_function: GreedyPolicyTrajectorySampler(model, reward_function)),
    ('blocks', lambda model, reward_function: EpsilonGreedyTrajectorySampler(model, reward_function, 0.5)),
    ('blocks', lambda model, reward_function: BeamSearchTrajectorySampler(model, reward_function, 8)),
    ('gripper', lambda model, reward_function: PolicyTrajectorySampler(model, reward_function)),
    ('gripper', lambda model, reward_function: BoltzmannTrajectorySampler(model, reward_function, 1.0)),
    ('gripper', lambda model, reward_function: GreedyPolicyTrajectorySampler(model, reward_function)),
    ('gripper', lambda model, reward_function: StateBoltzmannTrajectorySampler(model, reward_function, 1.0, 0.1, 10)),
    ('gripper', lambda model, reward_function: EpsilonGreedyTrajectorySampler(model, reward_function, 0.5)),
    ('gripper', lambda model, reward_function: BeamSearchTrajectorySampler(model, reward_function, 16))
])
def test_trajectory_sampler(domain_name: str, trajectory_sampler_creator: Callable[[ActionScalarModel, RewardFunction], TrajectorySampler]):
    domain_path = DATA_DIR / domain_name / 'domain.pddl'
    problem_path = DATA_DIR / domain_name / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    reward_function = GoalTransitionRewardFunction(1)
    trajectory_sampler = trajectory_sampler_creator(model, reward_function)
    trajectories = trajectory_sampler.sample([(problem.get_initial_state(), problem.get_goal_condition())], 10)
    assert isinstance(trajectories, list)
    assert len(trajectories) == 1
    trajectory = trajectories[0]
    assert isinstance(trajectory, Trajectory)
    assert trajectory.is_solution() or len(trajectory) == 10
    trajectory.validate()  # Performs asserts internally.


@pytest.mark.parametrize("domain_name", ['blocks-hard', 'gripper-hard'])
def test_state_hindsight(domain_name: str):
    domain_path = DATA_DIR / domain_name / 'domain.pddl'
    problem_path = DATA_DIR / domain_name / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    reward_function = ConstantRewardFunction(-1)
    trajectory_sampler = PolicyTrajectorySampler(model, reward_function)
    trajectory_refiner = StateHindsightTrajectoryRefiner(10)
    original_trajectories = trajectory_sampler.sample([(problem.get_initial_state(), problem.get_goal_condition())], 100)
    refined_trajectories = trajectory_refiner.refine(original_trajectories)
    assert len(refined_trajectories) < 10
    for refined_trajectory in refined_trajectories:
        assert refined_trajectory.is_solution()
        refined_trajectory.validate()


@pytest.mark.parametrize("domain_name", ['blocks-hard', 'gripper-hard'])
def test_propositional_hindsight(domain_name: str):
    domain_path = DATA_DIR / domain_name / 'domain.pddl'
    problem_path = DATA_DIR / domain_name / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    reward_function = ConstantRewardFunction(-1)
    trajectory_sampler = PolicyTrajectorySampler(model, reward_function)
    trajectory_refiner = PropositionalHindsightTrajectoryRefiner([problem], 10)
    original_trajectories = trajectory_sampler.sample([(problem.get_initial_state(), problem.get_goal_condition())], 100)
    refined_trajectories = trajectory_refiner.refine(original_trajectories)
    assert len(refined_trajectories) < 10
    for refined_trajectory in refined_trajectories:
        assert refined_trajectory.is_solution()
        refined_trajectory.validate()


@pytest.mark.parametrize("domain_name", ['blocks-hard', 'gripper-hard'])
def test_lifted_hindsight(domain_name: str):
    domain_path = DATA_DIR / domain_name / 'domain.pddl'
    problem_path = DATA_DIR / domain_name / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model = RGNNWrapper(domain)
    reward_function = ConstantRewardFunction(-1)
    trajectory_sampler = PolicyTrajectorySampler(model, reward_function)
    trajectory_refiner = LiftedHindsightTrajectoryRefiner([problem], 10)
    original_trajectories = trajectory_sampler.sample([(problem.get_initial_state(), problem.get_goal_condition())], 100)
    refined_trajectories = trajectory_refiner.refine(original_trajectories)
    assert len(refined_trajectories) < 10
    for refined_trajectory in refined_trajectories:
        assert refined_trajectory.is_solution()
        refined_trajectory.validate()


@pytest.mark.parametrize("domain_name", ['blocks', 'gripper'])
def test_off_policy_algorithm(domain_name: str):
    domain_path = DATA_DIR / domain_name / 'domain.pddl'
    problem_path = DATA_DIR / domain_name / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    problems = [problem]
    model = RGNNWrapper(domain)
    optimizer = torch.optim.Adam(model.parameters())
    lr_scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer)
    discount_factor = 0.999
    loss_function = DQNOptimization(model, optimizer, lr_scheduler, model, discount_factor, 10.0)
    reward_function = ConstantRewardFunction(-1)
    replay_buffer = PrioritizedReplayBuffer(100)
    trajectory_sampler = PolicyTrajectorySampler(model, reward_function)
    horizon = 100
    rollout_count = 2
    batch_size = 4
    train_steps = 8
    problem_sampler = UniformProblemSampler()
    initial_state_sampler = OriginalInitialStateSampler()
    goal_condition_sampler = OriginalGoalConditionSampler()
    trajectory_refiner = PropositionalHindsightTrajectoryRefiner(problems, 10)
    algorithm = OffPolicyAlgorithm(problems,
                                   loss_function,
                                   reward_function,
                                   replay_buffer,
                                   replay_buffer,
                                   trajectory_sampler,
                                   horizon,
                                   rollout_count,
                                   batch_size,
                                   train_steps,
                                   problem_sampler,
                                   initial_state_sampler,
                                   goal_condition_sampler,
                                   trajectory_refiner)
    algorithm.fit()
    algorithm.fit()
    algorithm.fit()

def test_algorithm_hooks():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    problems = [problem]
    model = RGNNWrapper(domain)
    optimizer = torch.optim.Adam(model.parameters())
    lr_scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer)
    discount_factor = 0.999
    loss_function = DQNOptimization(model, optimizer, lr_scheduler, model, discount_factor, 100.0)
    reward_function = ConstantRewardFunction(-1)
    replay_buffer = PrioritizedReplayBuffer(100)
    trajectory_sampler = PolicyTrajectorySampler(model, reward_function)
    horizon = 10
    rollout_count = 2
    batch_size = 4
    train_steps = 8
    problem_sampler = UniformProblemSampler()
    initial_state_sampler = OriginalInitialStateSampler()
    goal_condition_sampler = OriginalGoalConditionSampler()
    trajectory_refiner = PropositionalHindsightTrajectoryRefiner(problems, 10)
    algorithm = OffPolicyAlgorithm(problems,
                                   loss_function,
                                   reward_function,
                                   replay_buffer,
                                   replay_buffer,
                                   trajectory_sampler,
                                   horizon,
                                   rollout_count,
                                   batch_size,
                                   train_steps,
                                   problem_sampler,
                                   initial_state_sampler,
                                   goal_condition_sampler,
                                   trajectory_refiner)
    sample_problems: list[bool] = []
    sample_initial_states: list[bool] = []
    sample_goal_conditions: list[bool] = []
    sample_trajectories: list[bool] = []
    refine_trajectories: list[bool] = []
    pre_collect_experience: list[bool] = []
    post_collect_experience: list[bool] = []
    pre_optimize_model: list[bool] = []
    post_optimize_model: list[bool] = []
    train_step: list[bool] = []
    algorithm.register_on_sample_problems(lambda x: sample_problems.append(True))
    algorithm.register_on_sample_initial_states(lambda x: sample_initial_states.append(True))
    algorithm.register_on_sample_goal_conditions(lambda x: sample_goal_conditions.append(True))
    algorithm.register_on_sample_trajectories(lambda x: sample_trajectories.append(True))
    algorithm.register_on_refine_trajectories(lambda x: refine_trajectories.append(True))
    algorithm.register_on_pre_collect_experience(lambda: pre_collect_experience.append(True))
    algorithm.register_on_post_collect_experience(lambda: post_collect_experience.append(True))
    algorithm.register_on_pre_optimize_model(lambda: pre_optimize_model.append(True))
    algorithm.register_on_post_optimize_model(lambda: post_optimize_model.append(True))
    algorithm.register_on_train_step(lambda x, l: train_step.append(True))
    algorithm.fit()
    assert len(sample_problems) == 1
    assert len(sample_initial_states) == 1
    assert len(sample_goal_conditions) == 1
    assert len(sample_trajectories) == 1
    assert len(refine_trajectories) == 1
    assert len(pre_collect_experience) == 1
    assert len(post_collect_experience) == 1
    assert len(pre_optimize_model) == 1
    assert len(post_optimize_model) == 1
    assert (len(train_step) == train_steps) or (len(train_step) == 0)


def test_value_based_initial_state_sampler():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    problems: list[mm.Problem] = [problem] * 100
    model: ActionScalarModel = RGNNWrapper(domain)
    reward_function: RewardFunction = ConstantRewardFunction(-1)
    initial_state_sampler = TopValueInitialStateSampler(problems, model, reward_function, 0.5, 0.1, 10)
    # First, test without any additional states to the pool of initial states.
    # We expect to get the original initial state.
    sampled_initial_states_1 = initial_state_sampler.sample(problems)
    assert len(sampled_initial_states_1) == len(problems)
    assert all(sampled_initial_states_1[idx] == problem.get_initial_state() for idx, problem in enumerate(problems))
    # Second, test adding a bunch of states.
    trajectory_sampler: TrajectorySampler = BoltzmannTrajectorySampler(model, reward_function, 1.0)
    sampled_trajectory = trajectory_sampler.sample([(problem.get_initial_state(), problem.get_goal_condition()) for problem in problems], 100)
    for trajectory in sampled_trajectory:
        for transition in trajectory:
            initial_state_sampler.add_state(transition.current_state, transition.predicted_value)
    sampled_initial_states_2 = initial_state_sampler.sample(problems)
    assert len(sampled_initial_states_2) == len(problems)


def test_evaluation():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    problems: list[mm.Problem] = [problem]
    model: ActionScalarModel = RGNNWrapper(domain)
    reward_function: RewardFunction = ConstantRewardFunction(-1)
    trajectory_sampler: TrajectorySampler = GreedyPolicyTrajectorySampler(model, reward_function)
    criterias: list[EvaluationCriteria] = [CoverageCriteria(), LengthCriteria(False), TDErrorCriteria(False)]
    horizon: int = 100
    evaluation = PolicyEvaluation(problems, criterias, trajectory_sampler, horizon)
    best1, result1 = evaluation.evaluate()
    best2, result2 = evaluation.evaluate()
    assert best1
    assert not best2
    assert result1 == result2


def test_iw_subtrajectory_sampler():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    reward_function: RewardFunction = ConstantRewardFunction(-1)
    sampler = IWSubtrajectorySampler(reward_function, 2)
    subtrajectory = sampler.sample(problem.get_initial_state(), problem.get_goal_condition())
    assert isinstance(subtrajectory, Trajectory)
    subtrajectory.validate(False)
    assert len(subtrajectory) == 3


def test_trajectory_sampler_with_subtrajectories():
    domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
    problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
    domain = mm.Domain(domain_path)
    problem = mm.Problem(domain, problem_path)
    model: ActionScalarModel = RGNNWrapper(domain)
    reward_function: RewardFunction = ConstantRewardFunction(-1)
    subtrajectory_sampler = IWSubtrajectorySampler(reward_function, 2)
    trajectory_sampler = EpsilonGreedyTrajectorySampler(model, reward_function, 0.5, subtrajectory_sampler, 1.0)
    state_goals = [(problem.get_initial_state(), problem.get_goal_condition())]
    trajectories = trajectory_sampler.sample(state_goals, 100)
    assert isinstance(trajectories, list)
    assert len(trajectories) > 0
    trajectory = trajectories[0]
    trajectory.validate()
    assert trajectory.is_solution()
    assert len(trajectory) >= 3


# def test_lifted_ff():
#     domain_path = DATA_DIR / 'gripper' / 'domain.pddl'
#     problem_path = DATA_DIR / 'gripper' / 'problem.pddl'
#     domain = mm.Domain(domain_path)
#     problem = mm.Problem(domain, problem_path)
#     lifted_ff = LiftedFFHeuristic(problem, True)
#     initial_state = problem.get_initial_state()
#     initial_h_value = lifted_ff.compute_value(initial_state, False)
#     pass



import pymimir as mm
import torch

from abc import ABC, abstractmethod

from .models import ActionScalarModel
from .reward_functions import RewardFunction


class InitialStateSampler(ABC):
    """
    Abstract base class for sampling initial states.
    """

    @abstractmethod
    def sample(self, problems: list[mm.Problem]) -> list[mm.State]:
        """
        Sample initial states for the problems.

        Args:
            problems (list[mm.Problem]): A list of problem instances.

        Returns:
            list[mm.State]: A list of sampled initial states.
        """
        pass


class OriginalInitialStateSampler(InitialStateSampler):
    """
    Samples the initial state that is specified in the PDDL file.
    """

    def sample(self, problems: list[mm.Problem]) -> list[mm.State]:
        return [problem.get_initial_state() for problem in problems]


class TopValueInitialStateSampler(InitialStateSampler):
    """
    Samples the initial state based on their predicted Q-values with respect to the original goal.
    """

    def __init__(self, problems: list[mm.Problem], model: ActionScalarModel, reward_function: RewardFunction, sample_original_probability: float, temperature: float, max_buffer_size: int) -> None:
        super().__init__()
        assert isinstance(sample_original_probability, float), "Sample original probability must be a float."
        assert isinstance(temperature, float), "Temperature must be a float."
        assert isinstance(max_buffer_size, int), "Max buffer size must be an integer."
        assert sample_original_probability >= 0.0, "Sample original probability must not be negative."
        assert sample_original_probability <= 1.0, "Sample original probability must be at most 1."
        assert max_buffer_size > 0, "Max buffer size must be positive."
        self.problems = problems
        self.model = model
        self.reward_function = reward_function
        self.sample_original_probability = sample_original_probability
        self.temperature = temperature
        self.max_buffer_size = max_buffer_size
        self.state_buffers: dict[mm.Problem, list[mm.State]] = {problem: [problem.get_initial_state()] for problem in problems}
        self.value_buffers: dict[mm.Problem, list[float]] = {problem: [0.0] for problem in problems}

    def _update_state_values(self, problem: mm.Problem) -> None:
        with torch.no_grad():
            # TODO: This code can likely be cleaned up.
            self.model.eval()
            device = next(self.model.parameters()).device
            states = self.state_buffers[problem]
            values = self.value_buffers[problem]
            goal_condition = problem.get_goal_condition()
            state_goals = [(state, goal_condition) for state in states]
            qvalues_actions_list = self.model.forward(state_goals)
            rewards_list: list[torch.Tensor] = []
            for state, (_, actions) in zip(states, qvalues_actions_list):
                rewards = torch.tensor([self.reward_function(state, action, action.apply(state), goal_condition) for action in actions], dtype=torch.float, device=device)
                rewards_list.append(rewards)
            new_values = torch.stack([(q_values + rewards).max() for (q_values, _), rewards in zip(qvalues_actions_list, rewards_list)])
            assert len(values) == new_values.numel()
            for idx in range(len(values)):
                values[idx] = new_values[idx].item()

    def add_state(self, state: mm.State, value: float) -> None:
        def argmin(xs: list[float]) -> int:
            return min(range(len(xs)), key=lambda x: xs[x])
        states = self.state_buffers[state.get_problem()]
        values = self.value_buffers[state.get_problem()]
        if len(values) >= self.max_buffer_size:
            replace_idx = argmin(values)
            replace_value = values[replace_idx]
            if value > replace_value:
                states[replace_idx] = state
                values[replace_idx] = value
        else:
            states.append(state)
            values.append(value)

    def sample(self, problems: list[mm.Problem]) -> list[mm.State]:
        sampled_initial_states: list[mm.State] = []
        updated_problems: set[mm.Problem] = set()
        for problem in problems:
            sample_original_initial_state = torch.rand(1).item() < self.sample_original_probability
            if sample_original_initial_state:
                sampled_initial_states.append(problem.get_initial_state())
            else:
                if problem not in updated_problems:
                    self._update_state_values(problem)
                    updated_problems.add(problem)
                states = self.state_buffers[problem]
                values = self.value_buffers[problem]
                values_tensor = torch.tensor(values, dtype=torch.float) / self.temperature
                probabilities = values_tensor.softmax(dim=0)
                state_idx = probabilities.multinomial(num_samples=1)
                sampled_initial_states.append(states[state_idx])
        return sampled_initial_states
